--[[
Group O Matic, 2-21-08 (c). by Bryce Wilbur -- Anyone may use, modify, and/or redistribute this work, provided that this notice remains unmodified. This work may not be sold or included in any for-profit compilations.

1.7 (4-26-08)
-Added a "Restore Method" option to the Saves extras pane:
	-Smart means that if a person is missing from the saved raid, try to fill in their spot with any new people that have the same roles; Dumb behaves the old way
	-Separate means don't auto-arrange new people, instead try to fit them into the lower groups; Merge behaves the old way
	-Default is "Dumb - Merge" which is exactly how it has behaved up until now
-Clicking "Add MVP" while targeting a player from another server will now properly format the "Add MVP" input box with their full name (Name-Server)
-MVP names are no longer case-sensitive
-Minor tweaks to GUI elements such as the "Manage Saves" scroll frame
-Fixed a bug which would incorrectly report a bad user template after 5-10 raid restores in a row

1.6 (4-23-08)
-Attempt 1 to make the included PvE templates smarter about separating auras into different groups
-Attempt to make the auto-arrange not suck at arranging shaman (which is surprisingly hard x_x)
-Update to the talent interpreter, shamans' talents will map to more accurate roles now -- cached roles will still work, but won't benefit from this until the player's talents are actually scanned by GOM
-Added an "Extras" pane that allows you to rename/delete individual saved raids
-Added "limit" and "pass" commands for templates (see templates.lua)

1.5 (4-10-08)
-You can now select players that should be valued higher than others with the same roles by the auto-arrange (MVPs). This is enforced for all templates including the AB/EotS template (if there are more than 5 people at a node, MVPs will be placed in the full subgroup).
-Fixed a few bugs in the "Enter a name for this saved raid" dialog

1.4 (4-7-08)
-Trying to restore with the AB/EotS template selected will no longer deselect that template, it will just ignore it during the restore.
-You can now change the number of movement requests that GOM sends to the server at a time (default 1).
-Improved the logic that determines who to move next. Even if you leave the throttle set to 1, arranges will complete in (usually substantially) fewer steps.

1.3 (4-5-08)
-Finally fixed the talent scan bug that was classifying people as every role, or more than one.
-GOM now works for raid assistants as well as the raid leader.

1.2 (4-3-08)
-Added the ability to exclude groups from save/restore/auto
-No more this/argN stuff >.>

1.1 (4-2-08)
-Added an alternate PvE template for the auto-arranger
-Improved the way the template "group" command works, see templates.lua
-Beta AB/EotS dynamic arrange (group by subzone)
-Fixed some weirdness with the main window sometimes becoming unclickable, jumping to the bottom left, etc.

1.0 (3-25-08)
-Updated for WoW 2.4
-Fixed a bug that would sometimes cause the addon to read a person's talents wrong if their raid index was consecutive to another member of the same class' index
-Fixed a bug that caused the talent inspector to sometimes classify hybrid classes as everything at once
-Various other bug fixes.
-General code cleanup.
-Release!

0.0.6B (3-17-08)
-Greatly improved the logic that decides who to move next. It should no longer get stuck or give you "group is full" errors.
-Fixed the automatic throttling of the auto-arrange's group mover.

0.0.5B (3-16-08)
-The auto-arranger now supports custom user templates, for example to behave differently for BGs or specific boss fights. (see templates.lua if you want to create one)
-Included a sample AV template, which tries to put a healer in each group and separate classes out as much as possible.
-Various code improvements
-When arranging groups, the addon now makes sure that the previous move succeeded before it tries to make another move. This allows the auto-arrange to go faster without risk of problems if you are lagging.
-Fixed a bug that happened if you gave someone else leadership of the raid in the middle of an auto-arrange.
-Possible fix for the "Your group is full" bug in AV
-Added checks for combat, you can't auto-move people in combat.

0.0.3B (3-14-08)
-Expanded GUI.
-Added local talent scan.
-Fixed error when running the auto-arrange as a mage, warlock, rogue or hunter
-Fixed various small bugs
-/gom cancel will now work during talent collection as well as during the actual arrangement process
-GOM now works in Battleground raids

0.0.2B (3-10-08): Added mini-GUI pinned to the raid frame

0.0.1B (3-8-08): initial release


TODO:
-AB/EotS template...should it support the group exclusions?
-Extras pane to list roles, raid/guild/realm/all, scroll frame with names, filtered by primary role, or all (with primary role listed). right side can be the full list of all roles
]]

GrOM_Vars = {
	RolesByNameCache = {},
	SavedRaidLayouts = {},
	AlwaysSure = false,
	Silent = false,
	ShowGUI = true,
	LocalOverridesRemote = false, --old var name, what this==true does now is disable the direct talent check entirely
	userTemplateToUse = nil,
	expanded = false,
	exclude = {false, false, false, false, false, false, false, false},
	excludeSaves = false,
	autoThrottle = 1,
	MVPs = {},
	restoreMethod = "dumb-merge"
}

local function UpdateVars()
	if not GrOM_Vars.RolesByNameCache then
		GrOM_Vars.RolesByNameCache = {}
	end
	if not GrOM_Vars.SavedRaidLayouts then
		GrOM_Vars.SavedRaidLayouts = {}
	end
	if not GrOM_Vars.exclude then
		GrOM_Vars.exclude = {false, false, false, false, false, false, false, false}
	end
	if not GrOM_Vars.autoThrottle then
		GrOM_Vars.autoThrottle = 1
	end
	if not GrOM_Vars.MVPs then
		GrOM_Vars.MVPs = {}
	end
	if not GrOM_Vars.restoreMethod then
		GrOM_Vars.restoreMethod = "dumb-merge"
	end
end

GrOM.Version_String = "1.7.3"

--set this to 0.xxx for betas
GrOM.Version = 1.0001 -- this should only change if the talent check system is changed (because of a talent patch etc) or if there's some other reason for non-raid leaders to upgrade


GrOM.userTemplates = {}
GrOM.ArrangementFrameUpdateThrottle = 0.3
GrOM.ArrangementFrameBreakMaximum = 500
GrOM.TalentRequestDelayMinimum = 5.0
GrOM.TalentRequestDelayMaximum = 40.0
GrOM.resetErrorConditionThrottle = 6.0
GrOM.talentRequestTTL = 1.8 --time to wait before giving up on that person and moving on to the next

GrOM.ARRANGE_BREAK = 0

GrOM.InspectTainted = true
GrOM.SaveName = nil

local loc = GrOM.Localization

local roleIndex = {
	["MAIN_TANK"] = 1,
	["TANK"] = 2,
	["TANK_BUFFER"] = 3,
	["HEALER"] = 4,
	["MANA_BATTERY"] = 5,
	["HEAL_BUFFER"] = 6,
	["MELEE_BUFFER"] = 7,
	["RANGE_BUFFER"] = 8,
	["SPELL_BUFFER"] = 9,
	["MELEE_DPS"] = 10,
	["RANGE_DPS"] = 11,
	["SPELL_DPS"] = 12
}

--index maps onto roles above
local defaultClassRoles = {
	["WARRIOR"] = {false, true, false, false, false, false, true, false, false, true, false, false},
	["DRUID"] = {false, false, false, true, false, true, false, false, false, false, false, false},
	["SHAMAN"] = {false, false, false, true, true, false, true, true, true, false, false, false},
	["MAGE"] = {false, false, false, false, false, false, false, false, false, false, false, true},
	["WARLOCK"] = {false, false, true, false, false, false, false, false, false, false, false, true},
	["ROGUE"] = {false, false, false, false, false, false, false, false, false, true, false, false},
	["HUNTER"] = {false, false, false, false, false, false, false, false, false, false, true, false},
	["PRIEST"] = {false, false, false, true, false, false, false, false, false, false, false, false},
	["PALADIN"] = {false, false, true, true, false, false, false, false, false, false, false, false}
}

--if one of these is true, it is true for ALL members of that class, false means SOME OR ALL members of that class do not play that role
local universalClassRoles = {
	["WARRIOR"] = {false, false, false, false, false, false, true, false, false, false, false, false},
	["DRUID"] = {false, false, false, false, false, false, false, false, false, false, false, false},
	["SHAMAN"] = {false, false, false, false, false, false, false, false, false, false, false, false},
	["MAGE"] = {false, false, false, false, false, false, false, false, false, false, false, true},
	["WARLOCK"] = {false, false, false, false, false, false, false, false, false, false, false, true},
	["ROGUE"] = {false, false, false, false, false, false, false, false, false, true, false, false},
	["HUNTER"] = {false, false, false, false, false, false, false, false, false, false, true, false},
	["PRIEST"] = {false, false, false, false, false, false, false, false, false, false, false, false},
	["PALADIN"] = {false, false, false, false, false, false, true, true, true, false, false, false}
}

--these map talents to their closest PRIMARY function...left to right top to bottom by tree
local tank, heal, melee, range, spell = 1, 2, 3, 4, 5
local mainRolesWarrior = {
	{melee,tank,melee,tank,tank,tank,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee},
	{tank,melee,tank,melee,melee,tank,tank,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee},
	{tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank}
}
local mainRolesDruid = {
	{spell,tank,tank,spell,spell,spell,spell,heal,tank,spell,spell,spell,spell,spell,spell,spell,spell,spell,spell,spell,spell},
	{tank,tank,tank,tank,tank,tank,tank,melee,melee,melee,tank,melee,tank,heal,tank,tank,tank,melee,tank,tank,tank},
	{heal,tank,heal,heal,tank,heal,heal,tank,heal,heal,heal,heal,heal,heal,heal,heal,heal,heal,heal,heal}
}
local mainRolesShaman = {
	{spell,spell,spell,heal,spell,spell,spell,spell,spell,spell,spell,spell,spell,spell,spell,spell,spell,heal,spell,spell},
	{melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee}, --god damn this tree sucks...i'm not even going to bother reading all the descriptions -.-
	{heal,heal,heal,heal,heal,spell,heal,heal,heal,heal,heal,heal,heal,heal,heal,heal,heal,heal,heal,heal}
}
local mainRolesMage = false --always spell dps

local mainRolesWarlock = false --always spell dps

local mainRolesRogue = false --always melee dps

local mainRolesHunter = false --always ranged dps

local mainRolesPriest = {
	{heal,spell,heal,heal,heal,heal,heal,heal,heal,spell,spell,spell,heal,heal,spell,spell,spell,spell,spell,spell,spell,spell},
	{heal,heal,heal,spell,spell,heal,spell,heal,spell,heal,spell,heal,heal,heal,spell,heal,heal,heal,spell,heal,heal},
	{spell,spell,spell,spell,spell,spell,spell,spell,heal,spell,spell,spell,spell,spell,spell,spell,spell,spell,spell,spell,spell}
}
local mainRolesPaladin = {
	{melee,heal,heal,melee,heal,heal,heal,heal,heal,heal,tank,heal,heal,heal,heal,heal,heal,tank,heal,heal},
	{tank,tank,tank,heal,tank,heal,tank,tank,tank,tank,tank,spell,tank,tank,tank,tank,tank,tank,tank,tank,tank,tank},
	{melee,melee,melee,melee,tank,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,melee,tank,melee,melee}
}

local talentsToMainRolesByClass = {
	["WARRIOR"] = mainRolesWarrior,
	["DRUID"] = mainRolesDruid,
	["SHAMAN"] = mainRolesShaman,
	["MAGE"] = mainRolesMage,
	["WARLOCK"] = mainRolesWarlock,
	["ROGUE"] = mainRolesRogue,
	["HUNTER"] = mainRolesHunter,
	["PRIEST"] = mainRolesPriest,
	["PALADIN"] = mainRolesPaladin
}

local talentsToSpecialRolesByClass = {
	["WARRIOR"] = {
		{},
		{[8]="MELEE_BUFFER"}, -- Commanding Presence -- (default template) Only considered if the warrior isn't placed in a tanking group
		{}
	},
	["DRUID"] = {
		{[18]="SPELL_BUFFER"}, --Moonkin Form
		{[18]="MELEE_BUFFER"}, --Leader of the Pack -- (default template) Only considered if the druid isn't placed in a tanking group
		{[20]="TANK_BUFFER"}   --Tree of Life
	},
	["SHAMAN"] = { --shaman are really hard to work into this framework x_x
		{[5]="MELEE_BUFFER",[6]="SPELL_BUFFER",[7]="MELEE_BUFFER",[20]="SPELL_BUFFER"},  --These assignments aren't optimal, they need to be fixed.
		{[7]="MELEE_BUFFER",[12]="MELEE_BUFFER",[14]="MELEE_BUFFER"},	 --These assignments aren't optimal, they need to be fixed.
		{[16]="MANA_BATTERY"} --Mana Tide Totem
	},
	["MAGE"] = {
		{},
		{},
		{}
	},
	["WARLOCK"] = {
		{},
		{[2]="TANK_BUFFER"}, --Improved Imp -- (default template) These warlocks get priority for tank groups
		{}
	},
	["ROGUE"] = {
		{},
		{},
		{}
	},
	["HUNTER"] = {
		{[17]="RANGE_BUFFER"}, --Ferocious Inspiration -- (default template) Melee will get put in this group also if there are no available melee groups or other ranged for this group
		{[17]="RANGE_BUFFER"}, --Trueshot Aura -- (default template) Melee will get put in this group also if there are no available melee groups or other ranged for this group
		{}
	},
	["PRIEST"] = {
		{},
		{},
		{[13]="TANK_BUFFER",[21]="MANA_BATTERY"} --VE and VT -- (default template) VT is considered first for group placement
	},
	["PALADIN"] = {
		{},
		{[1]="TANK_BUFFER"}, --Improved Devotion Aura -- (default template) Only matters if the pally is not a tank
		{}
	},
}

local currentRaidGroupIndexByName = {}

local currentRaidGroupRolesByIndex = {}

--list of subgroups by raid index
local currentRaidGroupPlacement = {}

local optimalRaidGroupPlacement = {}

--used for the "start over" button if an auto-arrange is forced to stop (someone leaves the raid etc)
local lastAction = {false, ""} -- [1] is true if the last action was a raid restore, and if so, [2] will be the name of the saved raid

local outdatedCheck = false

local delayedActionInProgress = false

local lastRaidSetup = {}

local groupFullErr = false

local resetErrorTimer = 0

local ignoredExcluded = false

local movementQueue = {}

local lockedIDs = {}

local lockedGroups = {}

local queuedSetsByGroup = {5,5,5,5,5,5,5,5}

local function cout(msg)
	if(DEFAULT_CHAT_FRAME and not GrOM_Vars.Silent) then
		DEFAULT_CHAT_FRAME:AddMessage(loc.outputprefix .. msg, 0.0, 1.0, 0.0, 1.0)
	end
end

local function dout(msg)
	if DEFAULT_CHAT_FRAME then
		DEFAULT_CHAT_FRAME:AddMessage("GOM DEBUG: " .. msg, 1.0, 0.0, 0.0, 1.0)
	end
end

local function ValidateCheckBoxes()
	GOMaticOverrideRemote:SetChecked(GrOM_Vars.LocalOverridesRemote)
	GOMaticSilence:SetChecked(GrOM_Vars.Silent)
	GOMaticSure:SetChecked(GrOM_Vars.AlwaysSure)
	GOMaticExcludeSaves:SetChecked(GrOM_Vars.excludeSaves)
end

local function GetPVEorPVPUName(unit)
	local name, realm = UnitName(unit)

	if not name then
		return nil
	end

	if realm then
		name = name .. "-" .. realm
	end

	return name
end

--ty to Aquendyn for this function
local function GetCmd(msg)
	if msg then
 		local a,b,c=strfind(msg, "(%S+)");
 		if a then
 			return c, strsub(msg, b+2);
 		else
 			return "";
 		end
 	end
end

local function tremovebyval(tab, val)
	for k,v in pairs(tab) do
		if(v==val) then
			table.remove(tab, k)
			return true
		end
	end
	return false
end

local function BuildIndexByNameTable()
	for i=1, 40 do
		currentRaidGroupIndexByName[GetRaidRosterInfo(i) or "-"]=i
	end
end

local PingReplies = {}

local function PingFrame_OnEvent(self, event, arg1, arg2, arg3, arg4)
	if arg3 == "WHISPER" then
		if arg1 == "GOMPINGR" then
			table.insert(PingReplies, arg4)
			if arg2 ~= GrOM.Version_String then
				cout(arg4 .. ": " .. arg2 .. "**")
			else
				cout(arg4 .. ": " .. arg2)
			end
		end
	end
end

local PingFrameDelay = 3
local PingFrameStartTime = 0

local function PingFrame_OnUpdate()
	if PingFrameStartTime + PingFrameDelay < time() then
		if GetNumRaidMembers() > 1 then
			for i=1, 40 do
				local n = (GetRaidRosterInfo(i))
				if n then
					if not tremovebyval(PingReplies, n) then
						cout(n .. ": " .. loc.notinstalled)
					end
				end
			end
		end
		delayedActionInProgress = false
		GrOM.UpdateButtons()
		GOMatic_PingFrame:UnregisterEvent("CHAT_MSG_ADDON")
		GOMatic_PingFrame:SetScript("OnEvent", nil)
		GOMatic_PingFrame:SetScript("OnUpdate", nil)
		GOMatic_PingFrame = nil
	end
end

local function CreatePingFrame()
	PingReplies = {UnitName("player")}

	PingFrameStartTime = time()

	delayedActionInProgress = true
	GrOM.UpdateButtons()
	local f = CreateFrame("Frame","GOMatic_PingFrame")
	f:SetScript("OnEvent", PingFrame_OnEvent)
	f:SetScript("OnUpdate", PingFrame_OnUpdate)
	f:RegisterEvent("CHAT_MSG_ADDON")
	f:Show()
end

local function FindMemberID(n)
	if GetNumRaidMembers() > 1 then
		for i=1, 40 do
			if n == GetRaidRosterInfo(i) then
				return i
			end
		end
	end

	return false
end

function GrOM.RSCommandHandler(msg)
	if InCombatLockdown() then
		return
	end

	local cmd, sc = GetCmd(msg)

	if not (cmd and sc) then
		return
	end

	BuildIndexByNameTable()

	local i1, i2 = currentRaidGroupIndexByName[cmd], currentRaidGroupIndexByName[sc]

	if i1 and i2 then
		SwapRaidSubgroup(i1, i2)
	end
end

function GrOM.RGSCommandHandler(msg)
	if InCombatLockdown() then
		return
	end

	local cmd, sc = GetCmd(msg)

	if not (cmd and sc) then
		return
	end

	cmd = tonumber(cmd)
	sc = tonumber(sc)

	if not (cmd and sc) then
		return
	end

	if cmd == sc then
		return
	end

	if cmd > 8 or cmd < 1 or sc > 8 or sc < 1 then
		return
	end

	local g1, g2 = {}, {}

	for i = 1, 40 do
		local n,_,g = GetRaidRosterInfo(i)

		if n and g then
			if g == cmd then
				table.insert(g1, i)
			elseif g == sc then
				table.insert(g2, i)
			end
		end
	end

	for j = 1, 5 do
		if g1[j] and g2[j] then
			SwapRaidSubgroup(g1[j], g2[j])
		elseif g1[j] then
			SetRaidSubgroup(g1[j], sc)
		elseif g2[j] then
			SetRaidSubgroup(g2[j], cmd)
		end
	end
end

function GrOM.RMCommandHandler(msg)
	if InCombatLockdown() then
		return
	end

	local cmd, sc = GetCmd(msg)

	sc = tonumber(sc)

	if not (cmd and sc) then
		return
	end

	if sc < 1 or sc > 8 then
		return
	end

	BuildIndexByNameTable()

	local i = currentRaidGroupIndexByName[cmd]

	if i then
		SetRaidSubgroup(i, sc)
	end
end

function GrOM.Ping()
		if delayedActionInProgress then
			cout(loc.pleasewaiterr)
			return
		end

		CreatePingFrame()
		SendAddonMessage("GOMPING", tostring(GrOM.Version), "RAID")
end

function GrOM.Cancel()
	delayedActionInProgress = false
	GrOM.UpdateButtons()

	if GOMatic_ArrangementFrame and GOMatic_ArrangementFrame:IsVisible() then
		GOMatic_ArrangementFrame:SetScript("OnUpdate", nil)
		GOMatic_ArrangementFrame:SetScript("OnEvent", nil)
		GOMatic_ArrangementFrame = nil
		cout(loc.cancelled)
	end

	if GOMatic_TalentRequestFrame and GOMatic_TalentRequestFrame:IsVisible() then
		GOMatic_TalentRequestFrame:SetScript("OnEvent", nil)
		GOMatic_TalentRequestFrame:UnregisterEvent("CHAT_MSG_ADDON")
		GOMatic_TalentRequestFrame:UnregisterEvent("INSPECT_TALENT_READY")
		GOMatic_TalentRequestFrame:SetScript("OnUpdate", nil)
		GOMatic_TalentRequestFrame = nil
		cout(loc.cancelled)
	end

	if GOMatic_PingFrame and GOMatic_PingFrame:IsVisible() then
		GOMatic_PingFrame:UnregisterEvent("CHAT_MSG_ADDON")
		GOMatic_PingFrame:SetScript("OnEvent", nil)
		GOMatic_PingFrame:SetScript("OnUpdate", nil)
		GOMatic_PingFrame = nil
	end
end

function GrOM.SlashCommandHandler(msg)
	local cmd, sc = GetCmd(msg)

	if GetNumRaidMembers() < 2 then
		cout(loc.nopointerr)
		return
	end

	if cmd and string.lower(cmd) == loc.pingcmd then
		GrOM.Ping()
	elseif cmd == "roles" then
		if sc == "" then
			sc = nil
		end
		GrOM.PrintRoles(sc)
	elseif cmd and string.lower(cmd) == loc.guicmd then
		GrOM_Vars.ShowGUI = not GrOM_Vars.ShowGUI

		if GrOM_Vars.ShowGUI then
			cout(loc.guishown)
			if GetNumRaidMembers() > 1 then
				GOMatic:Show()
				GrOM.UpdateButtons()
			end
		else
			cout(loc.guihidden)
			GOMatic:Hide()
		end
	elseif cmd and string.lower(cmd) == loc.silentcmd then
		GrOM_Vars.Silent = not GrOM_Vars.Silent
		cout(loc.unsilenced)
	elseif cmd and string.lower(cmd) == loc.surecmd then
		GrOM_Vars.AlwaysSure = not GrOM_Vars.AlwaysSure
		cout(GrOM_Vars.AlwaysSure and loc.sureon or loc.sureoff)
	elseif cmd and string.lower(cmd) == loc.cancelcmd then
		GrOM.Cancel()
	elseif cmd and string.lower(cmd) == loc.savecmd then
		if delayedActionInProgress then
			cout(loc.pleasewaiterr)
			return
		end
		if sc and sc ~= "" then
			if GrOM_Vars.SavedRaidLayouts[sc] then
				GrOM.SaveName = sc
				StaticPopup_Show("GOMatic_DO_SAVE", sc)
			else
				GrOM.SaveRaid(sc)
			end
		else
			cout(loc.mustenternameerr)
		end
	elseif cmd and string.lower(cmd) == loc.restorecmd then
		if delayedActionInProgress then
			cout(loc.pleasewaiterr)
			return
		end
		if sc then
			local n = GrOM_Vars.SavedRaidLayouts[sc]
			if n then
				lastAction = {true, sc}
				GrOM.RestoreRaid(n)
			else
				cout(loc.mustenternameerr)
				cout(loc.thesearegoodnames)
				foreach(GrOM_Vars.SavedRaidLayouts, function(k,v) cout(k) end)
			end
		else
			cout(loc.mustenternameerr)
			cout(loc.thesearegoodnames)
			foreach(GrOM_Vars.SavedRaidLayouts, function(k,v) cout(k) end)
		end
	elseif cmd and string.lower(cmd) == loc.autocmd then
		if delayedActionInProgress then
			cout(loc.pleasewaiterr)
			return
		end
		lastAction = {false, ""}
		GrOM.StartArrange()
	else
		cout(loc.help)
	end

	ValidateCheckBoxes()
end

function GrOM.AutoClick()
	if delayedActionInProgress then
		cout(loc.pleasewaiterr)
		return
	end
	lastAction = {false, ""}
	GrOM.StartArrange()
end

function GrOM.RolesToString(roles)
	local msg = "{"

	for i=1, 11 do
		local x = (roles[i]) and "true, " or "false, "
		msg = msg .. x
	end
	local x = (roles[12]) and "true}" or "false}"
	msg = msg .. x

	return msg
end

local function GetMyRoles(u)
	if u == "raid0" then dout("x") return false end

	local mainRole = {0,0,0,0,0} --counter: tank/heal/melee/ranged/spell
	local myClass = select(2, UnitClass("player"))
	if u then
		myClass = select(2, UnitClass(u))
		u = true
	end
	local roles = CopyTable(universalClassRoles[myClass])
	local x = talentsToMainRolesByClass[myClass]

	if GetNumTalentTabs(u) ~= 3 then
		return false
	end

	if x then
		if GetNumTalents(1,u) ~= getn(x[1]) or GetNumTalents(2,u) ~= getn(x[2]) or GetNumTalents(3,u) ~= getn(x[3]) then
			return false
		end
	end

	if not x then
		--player is a pure class, just find special roles
		mainRole = nil

		for i=1, 3 do
			for j=1, GetNumTalents(i,u) do
				local _, _, _, _, rank, maxRank = GetTalentInfo(i,j,u)
				--this finds special roles, like "tank_buffer" for pallies with imp devo or locks with imp imp. generally these roles are simply "preferred" ie those players will get picked first to be in the tank groups
				if rank > 0 and rank == maxRank then

					local r = talentsToSpecialRolesByClass[myClass][i][j]
					if r then
						roles[roleIndex[r] ] = true
					end
				end
			end
		end
	else
		--player is a hybrid class, find main role and special roles
		for i=1, 3 do
			for j=1, GetNumTalents(i,u) do
				local _, _, _, _, rank, maxRank = GetTalentInfo(i,j,u)

				if rank > 0 and rank == maxRank then
					--this tries to figure out the primary role of the player by counting the number of tanking vs healing vs dps talents they have
					mainRole[x[i][j] ] = mainRole[x[i][j] ] + 1
					--this finds special roles, like "tank_buffer" for pallies with imp devo or locks with imp imp. generally these roles are simply "preferred" ie those players will get picked first to be in the tank groups
					local r = talentsToSpecialRolesByClass[myClass][i][j]
					if r then
						roles[roleIndex[r] ] = true
					end
				end
			end
		end
	end

	if mainRole then
		local roleHighest, myMainRole = 0, 1

		for a=1, 5 do
			if mainRole[a] > roleHighest then
				roleHighest = mainRole[a]
				myMainRole = a
			end
		end

		roles[roleIndex[({"TANK", "HEALER", "MELEE_DPS", "RANGE_DPS", "SPELL_DPS"})[myMainRole] ] ] = true
	end

	return roles
end

function GrOM.OnEvent(self, event, ...)
	if event == "VARIABLES_LOADED" then
		hooksecurefunc("NotifyInspect", function() GrOM.InspectTainted = true end) --make sure that noone else calls notifyinspect between the time we do and receiving talents
		UpdateVars()
		ValidateCheckBoxes()
		GrOM.UpdateButtons()
		GOMaticThrottleSlider:SetValue(GrOM_Vars.autoThrottle)
		GrOM.InitializeScrollFrame(GOMaticMVPScrollFrame)
		GrOM.UpdateMVPPane()
		GOMaticMVPScrollFrame:Hide()
		GrOM.InitializeScrollFrame(GOMaticSavesScrollFrame)
		GrOM.UpdateSavesPane()
		GOMaticSavesScrollFrame:Hide()
		cout(GrOM.Version_String .. loc.loaded)
	elseif event == "PLAYER_REGEN_DISABLED" then
		GrOM.UpdateButtons(1)
	elseif event == "PLAYER_REGEN_ENABLED" then
		GrOM.UpdateButtons(0)
	elseif event == "PLAYER_LEAVING_WORLD" then
		GrOM.CloseExtrasPane(true)
	elseif event == "RAID_ROSTER_UPDATE" then
		SendAddonMessage("GOMVER", tostring(GrOM.Version), "RAID")
		if GetNumRaidMembers() < 2 then
			self:Hide()
		else
			self:Show()
			GrOM.UpdateButtons()
		end
		if not GrOM_Vars.ShowGUI then
			self:Hide()
		end
	elseif event == "CHAT_MSG_ADDON" then
		local arg1, arg2, arg3, arg4 = select(1, ...)

		if arg3 == "RAID" then
			if arg1 == "GOMPING" then
				SendAddonMessage("GOMPINGR", GrOM.Version_String, "WHISPER", arg4)
			elseif arg1 == "GOMVER" then
				if tonumber(arg2) and tonumber(arg2) > GrOM.Version and GrOM.Version >= 1 then
					if not (outdatedCheck or InCombatLockdown() ) then
						StaticPopup_Show("GOMatic_OUTDATED")
						outdatedCheck = true
					end
				end
			elseif arg1 == "GOMTALENT" then
				local myRoles = GetMyRoles()
				if myRoles then
					SendAddonMessage("GOMTALENTR", GrOM.RolesToString(myRoles), "WHISPER", arg4)
				else
					cout(loc.talentserr)
				end
			end
		end
	end
end

function GrOM.PrintRoles(x)
	cout(GrOM.RolesToString(GetMyRoles(x)))
end

function GrOM.OnLoad(self)
	if GetNumRaidMembers() < 2 then
		self:Hide()
	end

	StaticPopupDialogs["GOMatic_DO_ARRANGE"] = {
		text = loc.doarrangedialogtext,
		button1 = loc.buttonok,
		button2 = loc.buttoncancel,
		OnAccept = function()
			GrOM.CreateArrangementFrame()
		end,
		timeout = 0,
		whileDead = 1,
		hideOnEscape = 1
	}

	StaticPopupDialogs["GOMatic_ERROR_TOOMANYFAILEDMOVES"] = {
		text = loc.toomanyfailedmovesdialogtext,
		button1 = loc.buttonok,
		timeout = 30,
		whileDead = 1,
		hideOnEscape = 1
	}

	StaticPopupDialogs["GOMatic_DO_SAVE"] = {
		text = loc.dosavedialogtext,
		button1 = loc.buttonok,
		button2 = loc.buttoncancel,
		OnAccept = function()
			GrOM.SaveRaid(GrOM.SaveName)
		end,
		timeout = 0,
		whileDead = 1,
		hideOnEscape = 1
	}

	StaticPopupDialogs["GOMatic_DO_RENAME"] = {
		text = loc.dosavedialogtext,
		button1 = loc.buttonok,
		button2 = loc.buttoncancel,
		OnAccept = function()
			GrOM.RenameSave(GrOM.SaveName, true)
		end,
		timeout = 0,
		whileDead = 1,
		hideOnEscape = 1
	}

	StaticPopupDialogs["GOMatic_AUTO_ABORT"] = {
		text = loc.autoabortdialogtext,
		button1 = loc.buttonstartover,
		button2 = loc.buttoncancel,
		OnAccept = function()
			GrOM.StartOver()
		end,
		timeout = 0,
		whileDead = 1,
		hideOnEscape = 1
	}

	StaticPopupDialogs["GOMatic_DELETE_ALL_SAVES"] = {
		text = loc.deletealldialogtext,
		button1 = loc.buttonok,
		button2 = loc.buttoncancel,
		OnAccept = function()
			GrOM.DeleteAllSavedRaids()
		end,
		timeout = 0,
		whileDead = 1,
		hideOnEscape = 1
	}

	StaticPopupDialogs["GOMATIC_CLEAR_MVP"] = {
		text = loc.deleteallmvpsdialogtext,
		button1 = loc.buttonok,
		button2 = loc.buttoncancel,
		OnAccept = function()
			GrOM.ClearMVPs()
		end,
		timeout = 0,
		whileDead = 1,
		hideOnEscape = 1
	}

	StaticPopupDialogs["GOMatic_OUTDATED"] = {
		text = loc.addonoutdateddialogtext,
		button1 = loc.buttonok,
		timeout = 30,
		whileDead = 1,
		hideOnEscape = 1
	}

	StaticPopupDialogs["GOMatic_NEW_SAVE"] = {
		text = loc.newsavedialogtext,
		button1 = loc.buttonok,
		button2 = loc.buttoncancel,
		hasEditBox = 1,
		maxLetters = 12,
		OnShow = function() --horrible...is "this" the only option for staticpopups?
			local editBox = getglobal(this:GetName().."EditBox")
			editBox:SetText("")
			editBox:SetFocus()
		end,
		OnAccept = function()
			local editBox = getglobal(this:GetParent():GetName().."EditBox")
			local name = editBox:GetText()
			name = string.gsub(name, " ", "")
			if not name or name == "" then
				cout(loc.mustenternameerr)
				return
			end
			GOMaticDropdown_DoSaveMenuItem(nil, name)
		end,
		EditBoxOnEnterPressed = function()
			local editBox = getglobal(this:GetParent():GetName().."EditBox")
			local name = editBox:GetText()
			name = string.gsub(name, " ", "")
			if not name or name == "" then
				cout(loc.mustenternameerr)
				return
			end
			GOMaticDropdown_DoSaveMenuItem(nil, name)
			this:GetParent():Hide()
		end,
		EditBoxOnEscapePressed = function()
			this:GetParent():Hide()
		end,
		timeout = 0,
		whileDead = 1,
		hideOnEscape = 1
	}

	StaticPopupDialogs["GOMATIC_RENAME_SAVE"] = {
		text = loc.renamesavedialogtext,
		button1 = loc.buttonok,
		button2 = loc.buttoncancel,
		hasEditBox = 1,
		maxLetters = 12,
		OnShow = function()
			local editBox = getglobal(this:GetName().."EditBox")
			editBox:SetText("")
			editBox:SetFocus()
		end,
		OnAccept = function()
			local editBox = getglobal(this:GetParent():GetName().."EditBox")
			local name = editBox:GetText()
			name = string.gsub(name, " ", "")
			if not name or name == "" then
				cout(loc.mustenternameerr)
				return
			end
			GrOM.RenameSave(name)
		end,
		EditBoxOnEnterPressed = function()
			local editBox = getglobal(this:GetParent():GetName().."EditBox")
			local name = editBox:GetText()
			name = string.gsub(name, " ", "")
			if not name or name == "" then
				cout(loc.mustenternameerr)
				return
			end
			GrOM.RenameSave(name)
			this:GetParent():Hide()
		end,
		EditBoxOnEscapePressed = function()
			this:GetParent():Hide()
		end,
		timeout = 0,
		whileDead = 1,
		hideOnEscape = 1
	}

	StaticPopupDialogs["GOMATIC_ADD_MVP"] = {
		text = loc.addmvpdialogtext,
		button1 = loc.buttonok,
		button2 = loc.buttoncancel,
		hasEditBox = 1,
		maxLetters = 14,
		OnShow = function()
			local editBox = getglobal(this:GetName().."EditBox")
			editBox:SetText(GetPVEorPVPUName("target") or "")
			editBox:SetFocus()
		end,
		OnAccept = function()
			local editBox = getglobal(this:GetParent():GetName().."EditBox")
			local name = editBox:GetText()
			name = string.gsub(name, " ", "")
			if not name or name == "" then
				cout(loc.mustenternameerr)
				return
			end
			GrOM.AddMVP(name)
		end,
		EditBoxOnEnterPressed = function()
			local editBox = getglobal(this:GetParent():GetName().."EditBox")
			local name = editBox:GetText()
			name = string.gsub(name, " ", "")
			if not name or name == "" then
				cout(loc.mustenternameerr)
				return
			end
			GrOM.AddMVP(name)
			this:GetParent():Hide()
		end,
		EditBoxOnEscapePressed = function()
			this:GetParent():Hide()
		end,
		timeout = 0,
		whileDead = 1,
		hideOnEscape = 1
	}

	GOMaticSave:SetText(loc.savebutton)
	GOMaticAuto:SetText(loc.autobutton)
	GOMaticHide:SetText(loc.hidebutton)
	GOMaticMore:SetText(loc.morebutton)
	GOMaticLess:SetText(loc.lessbutton)
	GOMaticPing:SetText(loc.pingbutton)
	GOMaticCancel:SetText(loc.buttoncancel)
	GOMaticAddMVP:SetText(loc.addmvpbutton)
	GOMaticRestore:SetText(loc.restorebutton)
	GOMaticSureText:SetText(loc.surebutton)
	GOMaticDeleteMVP:SetText(loc.delmvpbutton)
	GOMaticDemoteMVP:SetText(loc.demotemvpbutton)
	GOMaticTitleText:SetText("GOM:")	
	GOMaticPromoteMVP:SetText(loc.promotemvpbutton)
	GOMaticDeleteSave:SetText(loc.delsavebutton)
	GOMaticRenameSave:SetText(loc.renamesavebutton)
	GOMaticExcludeText:SetText(loc.excludetext)
	GOMaticSilenceText:SetText(loc.silencebutton)
	GOMaticThrottleText:SetText(loc.throttletext:format(GrOM_Vars.autoThrottle))
	GOMaticTemplateText:SetText(loc.templatetext)
	GOMaticDeleteAllMVPs:SetText(loc.dellallmvpbutton)
	GOMaticExcludeSavesText:SetText(loc.excludesavebutton)
	GOMaticRestoreMethodText:SetText(loc.restoremethodtext)
	GOMaticOverrideRemoteText:SetText(loc.overridebutton)

	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("VARIABLES_LOADED")
	self:RegisterEvent("CHAT_MSG_ADDON")
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PLAYER_LEAVING_WORLD")

	SlashCmdList["GROUPOMATIC"] = GrOM.SlashCommandHandler
	SLASH_GROUPOMATIC1 = "/groupomatic"
	SLASH_GROUPOMATIC2 = "/gom"

	SlashCmdList["GROUPOMATIC_RAID_MOVE"] = GrOM.RMCommandHandler
	SLASH_GROUPOMATIC_RAID_MOVE1 = "/raidmove"
	SLASH_GROUPOMATIC_RAID_MOVE2 = "/rm"

	SlashCmdList["GROUPOMATIC_RAID_SWAP"] = GrOM.RSCommandHandler
	SLASH_GROUPOMATIC_RAID_SWAP1 = "/raidswap"
	SLASH_GROUPOMATIC_RAID_SWAP2 = "/rs"

	SlashCmdList["GROUPOMATIC_RAID_GROUP_SWAP"] = GrOM.RGSCommandHandler
	SLASH_GROUPOMATIC_RAID_GROUP_SWAP1 = "/raidgroupswap"
	SLASH_GROUPOMATIC_RAID_GROUP_SWAP2 = "/rgs"
end

local function GetAllRaidIDs()
	local allIDs = false
	local ignoreIDs = {}

	for i =1, getn(GrOM_Vars.MVPs) do
		local mvp = strlower(GrOM_Vars.MVPs[i])
		for j = 1, 40 do
			local n,_,g = GetRaidRosterInfo(j)
			if n and g and strlower(n) == mvp then
				if not allIDs then
					allIDs = {[1]=j}
				else
					table.insert(allIDs, j)
				end
				table.insert(ignoreIDs, j)
			end
		end
	end

	local function FindIgnored(n)
		for q = 1, getn(ignoreIDs) do
			if ignoreIDs[q] == n then
				return true
			end
		end

		return false
	end

	for i=1, 40 do
		if not FindIgnored(i) then
			local n,_,g = GetRaidRosterInfo(i)
			if n and g then
				if not allIDs then
					allIDs = {[1]=i}
				else
					table.insert(allIDs, i)
				end
			end
		end
	end

	return allIDs
end

local function FillInEmptyRoles(displayIfAnyWereEmpty) --if this is true DON'T display if any were empty
	local wereEmpty = false

	for i=1, 40 do
		if not currentRaidGroupRolesByIndex[i] and UnitExists("raid"..i) then
			currentRaidGroupRolesByIndex[i] = GrOM_Vars.RolesByNameCache[GetPVEorPVPUName("raid"..i)] or defaultClassRoles[select(2, UnitClass("raid"..i))]
			wereEmpty = true
		end
	end

	if wereEmpty and not displayIfAnyWereEmpty then
		cout(loc.filledinunknownroleswithdefaultsrunping)
	end
end

local function ArrangeManager()
	delayedActionInProgress = false
	GrOM.UpdateButtons()

	optimalRaidGroupPlacement = {}

	local display = (GrOM_Vars.userTemplateToUse == "AB_EOTS_AUTO_AKRYN")

	FillInEmptyRoles(display)

	ignoredExcluded = false
	local optimalSetup = GrOM.Arrange()

	if not optimalSetup then
		return
	end

	local g = 1

	local group = optimalSetup[g]

	while group and group ~= {} do
		for i=1, 5 do
			if group[i] then
				optimalRaidGroupPlacement[group[i] ] = g
			end
		end
		g = g + 1
		group = optimalSetup[g]

	end

	if not GrOM_Vars.AlwaysSure then
		StaticPopup_Show("GOMatic_DO_ARRANGE")
	else
		GrOM.CreateArrangementFrame()
	end
end

function GrOM.StartOver()
	if lastAction[1] then
		local n = GrOM_Vars.SavedRaidLayouts[lastAction[2]]
		if n then
			GrOM.RestoreRaid(n)
		end
	else
		GrOM.StartArrange()
	end
end

local function BuildCurrentPlacementTable()
	currentRaidGroupPlacement = {}

	local ct, tt = {}, {}

	for i=1, 40 do
		local n,_,x = GetRaidRosterInfo(i)
		if not n then return end
		if x and x > 0 and x < 9 then
			currentRaidGroupPlacement[i] = x
			if ct[1] then
				for j=1, getn(ct) do
					table.insert(tt, ct[j])
				end

				ct = {}
			end
		else
			currentRaidGroupPlacement[i] = nil
			table.insert(ct, i)
		end
	end

	local breaker = 10

	while breaker > 0 and tt[1] do
		breaker = breaker - 1

		local k = 0
		while k < getn(tt) do
			k = k + 1

			local _,_,a = GetRaidRosterInfo(i)

			if a and a > 0 and a < 9 then
				currentRaidGroupPlacement[k] = a
				table.remove(tt, k)
				k = k - 1
			end
		end
	end

	if tt[1] then
		return -1
	end
end

local function FindFirstEmptyOptimalSlot()
	local slotNumbers = {0,0,0,0,0,0,0,0}

	for i = 1, 40 do
		if optimalRaidGroupPlacement[i] and optimalRaidGroupPlacement[i] < 9 and optimalRaidGroupPlacement[i] > 0 then
			slotNumbers[optimalRaidGroupPlacement[i] ] = slotNumbers[optimalRaidGroupPlacement[i] ] + 1
		end
	end

	for j = 1, 8 do
		if slotNumbers[j] < 5 and not GrOM_Vars.exclude[j] then
			return j
		end
	end

	for j = 1, 8 do
		if slotNumbers[j] < 5 then
			return j
		end
	end

	return false
end

local function RepairOptimalPlacementTable()
	for i = 1, 40 do
		if not currentRaidGroupPlacement[i]  then
			optimalRaidGroupPlacement[i] = false
		end

		if currentRaidGroupPlacement[i] and currentRaidGroupPlacement[i] < 9 and currentRaidGroupPlacement[i] > 0 then
			if (not optimalRaidGroupPlacement[i]) or (optimalRaidGroupPlacement[i] < 1 or optimalRaidGroupPlacement[i] > 8) then
				optimalRaidGroupPlacement[i] = FindFirstEmptyOptimalSlot()
				if not optimalRaidGroupPlacement[i] then
					GrOM.Cancel()
					cout(loc.memberaddederr)
				end
			end
		end
	end
end

function GrOM.CreateArrangementFrame()
	groupFullErr = false

	lastRaidSetup = {}

	resetErrorTimer = 0

	BuildCurrentPlacementTable()

	GOMatic_ArrangementFrameLastUpdate = GetTime()
	GrOM.ArrangementFrameBreakCounter = 0

	delayedActionInProgress = true
	GrOM.UpdateButtons()
	local f = CreateFrame("Frame","GOMatic_ArrangementFrame")
	f:SetScript("OnEvent", GOMatic_ArrangementFrame_OnEvent)
	f:SetScript("OnUpdate", GOMatic_ArrangementFrame_OnUpdate )
	f:RegisterEvent("CHAT_MSG_SYSTEM")
	f:Show()
end

local function IsValidSpec(spec)
	if type(spec) ~= "table" then
		return false
	end

	if getn(spec) ~= 12 then
		return false
	end

	for i=1, 12 do
		if type(spec[i])~= "boolean" then
			return false
		end
	end

	return spec[roleIndex["TANK"]] or spec[roleIndex["HEALER"]] or spec[roleIndex["MELEE_DPS"]] or spec[roleIndex["RANGE_DPS"]] or spec[roleIndex["SPELL_DPS"]]
end

local function PlayerRespec(s1, s2)
	local msg = loc.playerhasrespecced

	if s2[roleIndex["TANK"]] and not s1[roleIndex["TANK"]] then
		msg = msg .. loc.spectank
	elseif s2[roleIndex["HEALER"]] and not s1[roleIndex["HEALER"]] then
		msg = msg .. loc.spechealer
	elseif (s2[roleIndex["MELEE_DPS"]] or s2[roleIndex["RANGE_DPS"]] or s2[roleIndex["SPELL_DPS"]]) and not (s1[roleIndex["MELEE_DPS"]] or s1[roleIndex["RANGE_DPS"]] or s1[roleIndex["SPELL_DPS"]]) then
		msg = msg .. loc.specdps
	else
		return nil
	end

	return msg
end

local function GetNumAssignedMembers(optimalSetup, countExcluded)
	local num = 0

	for i = 1, 8 do
		if countExcluded or (not GrOM_Vars.exclude[i]) then
			num = num + getn(optimalSetup[i])
		end
	end

	return num
end

local function GetMaxGroups(unassignedMembers, optimalSetup)
	local num, count = 0, getn(unassignedMembers) + GetNumAssignedMembers(optimalSetup, false)

	for i = 1, 8 do
		if not GrOM_Vars.exclude[i] then
			count = count - 5
			if count < 1 then
				return i
			end
		end
	end

	return 8
end

--[[local function GetMaxGroups(unassignedMembers)
	local x = (GetNumRaidMembers() + FindNumExcludedEmptySpaces(unassignedMembers)) / 5
	return math.ceil(x)
end]]

function GrOM.RemoveFromCache(n)
	GrOM_Vars.RolesByNameCache[n] = nil
end

local raidInspectionIndex = 1
local doneInspecting = false

function GrOM.TalentRequest_OnEvent(self, event, ...)
	if event == "CHAT_MSG_ADDON" then
		local arg1, arg2, arg3, arg4 = select(1, ...)
		if arg3 == "WHISPER" and arg1 == "GOMTALENTR" then
			GrOM.SpecRecieved = {}
			RunScript("GrOM.SpecRecieved = " .. arg2)
			local spec = GrOM.SpecRecieved

			if not IsValidSpec(spec) then
				return
			end

			if currentRaidGroupRolesByIndex[currentRaidGroupIndexByName[arg4] ] then
				return
			end

			currentRaidGroupRolesByIndex[currentRaidGroupIndexByName[arg4] ] = spec

			if GrOM_Vars.RolesByNameCache[arg4] then
				local respec = PlayerRespec(GrOM_Vars.RolesByNameCache[arg4], spec)
				if respec then
					cout(string.format(respec, arg4))
				end
			end

			GrOM_Vars.RolesByNameCache[arg4] = spec
		end
	elseif event == "INSPECT_TALENT_READY" then
		if GrOM_Vars.LocalOverridesRemote then
			return
		end
		local u = "raid" .. raidInspectionIndex
		if GrOM.InspectTainted then

			if (not UnitIsDead("player") ) and UnitExists(u) and UnitInRange(u) and (not UnitIsUnit(u, "player") ) then
				NotifyInspect(u)
				GrOM.inspectSent = GetTime()
			else
				while raidInspectionIndex < 40 do
					raidInspectionIndex = raidInspectionIndex + 1
					u = "raid" .. raidInspectionIndex

					if (not UnitIsDead("player") ) and UnitExists(u) and UnitInRange(u) and (not UnitIsUnit(u, "player") ) then
						GOMatic_TalentRequestFrame:RegisterEvent("INSPECT_TALENT_READY")
						NotifyInspect(u)
						GrOM.inspectSent = GetTime()
						GrOM.InspectTainted = false
						return
					end
				end
			end

			GrOM.InspectTainted = false
			return
		end

		GOMatic_TalentRequestFrame:UnregisterEvent("INSPECT_TALENT_READY")

		if GetTime() - GrOM.TalentRequestStartTime > 10 then
			if raidInspectionIndex == 10 then
				cout("25%")
			elseif raidInspectionIndex == 20 then
				cout("50%")
			elseif raidInspectionIndex == 30 then
				cout("75%")
			end
		end

		local spec = GetMyRoles(u)
		local name = GetPVEorPVPUName(u)

		if spec and IsValidSpec(spec) then
			if currentRaidGroupIndexByName[name] then
				currentRaidGroupRolesByIndex[currentRaidGroupIndexByName[name] ] = spec

				if GrOM_Vars.RolesByNameCache[name] then
					local respec = PlayerRespec(GrOM_Vars.RolesByNameCache[name], spec)
					if respec then
						cout(string.format(respec, name))
					end
				end

				GrOM_Vars.RolesByNameCache[name] = spec
			end
		end

		raidInspectionIndex = raidInspectionIndex + 1

		if raidInspectionIndex == 41 then
			raidInspectionIndex = 0
			doneInspecting = true
			return
		end

		u = "raid" .. raidInspectionIndex

		if (not UnitIsDead("player") ) and UnitExists(u) and UnitInRange(u) and (not UnitIsUnit(u, "player") ) then
			GOMatic_TalentRequestFrame:RegisterEvent("INSPECT_TALENT_READY")
			NotifyInspect(u)
			GrOM.inspectSent = GetTime()
			GrOM.InspectTainted = false
		elseif raidInspectionIndex < 40 then
			while raidInspectionIndex < 40 and raidInspectionIndex > 0 do
				raidInspectionIndex = raidInspectionIndex + 1
				u = "raid" .. raidInspectionIndex

				if (not UnitIsDead("player") ) and UnitExists(u) and UnitInRange(u) and (not UnitIsUnit(u, "player") ) then
					GOMatic_TalentRequestFrame:RegisterEvent("INSPECT_TALENT_READY")
					NotifyInspect(u)
					GrOM.inspectSent = GetTime()
					GrOM.InspectTainted = false
					return
				end
			end

			raidInspectionIndex = 0
			doneInspecting = true
			return
		end
	end
end

local function TalentRequest_OnUpdate()
	if GrOM.inspectSent + GrOM.talentRequestTTL < GetTime() then
		if raidInspectionIndex < 40 then
			while raidInspectionIndex < 40 and raidInspectionIndex > 0 do
				raidInspectionIndex = raidInspectionIndex + 1
				u = "raid" .. raidInspectionIndex

				if (not UnitIsDead("player") ) and UnitExists(u) and UnitInRange(u) and (not UnitIsUnit(u, "player") ) then
					GOMatic_TalentRequestFrame:RegisterEvent("INSPECT_TALENT_READY")
					NotifyInspect(u)
					GrOM.InspectTainted = false
					GrOM.inspectSent = GetTime()
					return
				end
			end

			raidInspectionIndex = 0
			doneInspecting = true
		end
	end

	if GrOM.TalentRequestStartTime + GrOM.TalentRequestDelayMinimum < GetTime() or GrOM_Vars.userTemplateToUse == "AB_EOTS_AUTO_AKRYN" then
		if not (GrOM_Vars.LocalOverridesRemote or GrOM_Vars.userTemplateToUse == "AB_EOTS_AUTO_AKRYN") then
			 if not (doneInspecting or GrOM.TalentRequestStartTime + GrOM.TalentRequestDelayMaximum < GetTime()  ) then
			 	return
			 end
		end

		ArrangeManager()
		GOMatic_TalentRequestFrame:SetScript("OnEvent", nil)
		GOMatic_TalentRequestFrame:UnregisterEvent("CHAT_MSG_ADDON")
		GOMatic_TalentRequestFrame:UnregisterEvent("INSPECT_TALENT_READY")
		GOMatic_TalentRequestFrame:SetScript("OnUpdate", nil)
		GOMatic_TalentRequestFrame = nil
	end
end

local function CreateTalentRequestFrame()
	GrOM.inspectSent = GetTime()
	raidInspectionIndex = 1
	doneInspecting = false
	GrOM.TalentRequestStartTime = GetTime()

	delayedActionInProgress = true
	GrOM.UpdateButtons()
	local f = CreateFrame("Frame","GOMatic_TalentRequestFrame")
	f:SetScript("OnUpdate", TalentRequest_OnUpdate)
	f:SetScript("OnEvent", GrOM.TalentRequest_OnEvent)
	f:RegisterEvent("CHAT_MSG_ADDON")
	f:Show()
	SendAddonMessage("GOMTALENT", tostring(GrOM.Version), "RAID")


	if not GrOM_Vars.LocalOverridesRemote then
		f:RegisterEvent("INSPECT_TALENT_READY")
		local u = ""
		for i=1, 40 do
			u = "raid" .. i
			if (not UnitIsDead("player") ) and UnitExists(u) and UnitInRange(u) and (not UnitIsUnit(u, "player") ) then
				raidInspectionIndex = i
				GOMatic_TalentRequestFrame:RegisterEvent("INSPECT_TALENT_READY")
				NotifyInspect(u)
				GrOM.inspectSent = GetTime()
				GrOM.InspectTainted = false
				return
			end
		end
		raidInspectionIndex = 0
		doneInspecting = true
	end
end

local function BuildRolesByIndexTable()
	for i=1, 40 do
		currentRaidGroupRolesByIndex[i] = false
	end
end

local function FindWorkingGroup(optimalSetup, unassignedMembers)
	local group = -1

	if GetNumRaidMembers() == 0 then
		return -1
	end

	local maxG = GetMaxGroups(unassignedMembers, optimalSetup)

	if maxG > 8 or maxG < 1 then
		return -2
	end

	for i=1, maxG do
		if (not optimalSetup[i][1]) then
			if not GrOM_Vars.exclude[i] then
				group = i
				break
			end
		end
	end

	if group == -1 then
		for i=1, 8 do
			if optimalSetup[i][1] then
				if getn(optimalSetup[i]) < 5 then
					if not GrOM_Vars.exclude[i] then
						group = i
						break
					end
				end
			end
		end
	end

	return group
end

function GrOM.StartArrange()
	if not (IsRaidLeader() or IsRaidOfficer()) then
		cout(loc.notraidleadererror)
		return
	end

	GrOM.ARRANGE_BREAK = 0

	BuildIndexByNameTable()
	BuildRolesByIndexTable()
	local myRoles = GetMyRoles()
	if not myRoles then
		cout(loc.talentserr)
		myRoles = defaultClassRoles[select(2,UnitClass("player"))]
	end
	currentRaidGroupRolesByIndex[currentRaidGroupIndexByName[UnitName("player")] ] = myRoles

	if not (GrOM_Vars.userTemplateToUse == "AB_EOTS_AUTO_AKRYN") then
		cout(loc.requestingtalents)
	end

	CreateTalentRequestFrame()
end

--thanks to Mikk for this
local function tcount(tab)
	local n=0

	for _ in pairs(tab) do
		n=n+1
	end

	return n
end


local function IsGroupFull(group)
	local n = 0

	for i=1, 40 do
		if currentRaidGroupPlacement[i] and currentRaidGroupPlacement[i] == group then
			n = n + 1
		end
	end

	return n > 4
end

local function IDIsLocked(i)
	for k = 1, getn(lockedIDs) do
		if lockedIDs[k] == i then
			return true
		end
	end

	return false
end

local function groupIsLocked(g)
	for k = 1, getn(lockedGroups) do
		if lockedGroups[k] == g then
			return true
		end
	end

	return false
end

local function FindIDToMoveToGroup(group, inGroup, requireInGroup)
	local setGroup = false

	if inGroup then
		if inGroup < 1 or inGroup > 8 then
			if inGroup == -17 then
				inGroup = nil
				setGroup = true
			else
				return nil
			end
		end
	end

	local match = nil

	for i=1, 40 do
		if not IDIsLocked(i) then
			local currentGroup = currentRaidGroupPlacement[i]
			local destGroup = optimalRaidGroupPlacement[i]

			if not (setGroup and groupIsLocked(destGroup)) then
				if currentGroup and currentGroup > 0 and currentGroup < 9 and destGroup and destGroup ~= currentGroup and destGroup > 0 and destGroup < 9 and destGroup == group then
					if inGroup then
						if currentGroup == inGroup then
							return i
						else
							match = i
						end
					else
						return i
					end
				end
			end
		end
	end

	if requireInGroup then
		return false
	end

	return match
end

local function FindEmptySlot()
	for i=1, 8 do
		if not IsGroupFull(i) then
			local x = FindIDToMoveToGroup(i, -17)
			if x then
				return i, x
			end
		end
	end

	return -1, false
end

local function ValidateTemplate(template)

	if type(template) ~= "table" then
		return false
	end

	if getn(template) ~= 3 then
		return false
	end

	if type(template[1]) ~= "string" then
		return false
	end

	if template[1] == "" then
		return false
	end

	if type(template[2]) ~= "table" then
		return false
	end

	if type(template[3]) ~= "table" then
		return false
	end

	local fail = false

	foreachi(template[2], function(i,v) if not ( (type(v)=="string") or (type(v)=="table") ) then fail = true end end)
	foreachi(template[3], function(i,v) if not ( (type(v)=="string") or (type(v)=="table") ) then fail = true end end)

	return not fail
end

local function ValidateGroups(...)
	local OKGroups = {}

	for i = 1, select("#", ...) do
		table.insert(OKGroups, tonumber(select(i, ...)))
	end

	if getn(OKGroups) < 1 or getn(OKGroups) > 8 then
		return false
	end

	for i = 1, getn(OKGroups) do
		local x = OKGroups[i]

		if x then
			if not x or x < 1 or x > 8 then
				return false
			end
		end
	end

	return OKGroups
end

local function myTFind(tab, i)
	for j = 1, getn(tab) do
		if tab[j] == i then
			return true
		end
	end

	return false
end

--class, prefSpec, requirePref, ignoreSpec, loopWhile
local function ValidateTemplateTable(t)
	if type(t) ~= "table" then
		return false
	end

	if getn(t) < 1 or getn(t) > 5 then
		return false
	end

	if t[1] then
		if type(t[1]) ~= "string" then
			return false
		end

		if not defaultClassRoles[t[1] ] then
			return false
		end
	end

	if not t[2] then
		return false
	end

	if type(t[2]) ~= "string" then
		return false
	end

	if not roleIndex[t[2] ]  then
		return false
	end

	if t[4] and (not roleIndex[t[4] ] ) then
		return false
	end

	return true
end

function GrOM.DisplayCurrAndDest()
	for i=1, 40 do
		currentGroup = currentRaidGroupPlacement[i]
		destGroup = optimalRaidGroupPlacement[i]
		local msg = ""
		if currentGroup then msg = currentGroup end
		if destGroup then msg = msg .. "     " .. destGroup end

		dout(msg)
	end
end

local function FindUnmovedID(ignoreGroup)
	local raidIDToMove, currentGroup, destGroup = nil, nil, nil
	local retCurr, retDest = nil, nil

	local i = 41
	while i > 1 do
		i = i - 1
		if not IDIsLocked(i) then
			currentGroup = currentRaidGroupPlacement[i]
			destGroup = optimalRaidGroupPlacement[i]


			if currentGroup and currentGroup > 0 and currentGroup < 9 and (not (ignoreGroup and currentGroup == ignoreGroup)) and destGroup and destGroup ~= currentGroup and destGroup > 0 and destGroup < 9 then
				raidIDToMove = i
				retCurr = currentGroup
				retDest = destGroup
				if FindIDToMoveToGroup(currentGroup, destGroup, true) then
					return raidIDToMove, retCurr, retDest
				end
			end
		end
	end

	return raidIDToMove, retCurr, retDest
end

--return 41 if a complete match, raid id if a match except for spec AND requirePref is false, or false if not a good match
local function DoesIDMatchClassAndSpec(ID, class, prefSpec, requirePref, ignoreSpec)
	if type(currentRaidGroupRolesByIndex[ID]) ~= "table" or not IsValidSpec(currentRaidGroupRolesByIndex[ID]) then
		return false
	end

	local realSpec, realClass = currentRaidGroupRolesByIndex[ID], select(2, UnitClass("raid"..ID))

	local doesMatch = 41

	if class and realClass ~= class then
		return false
	end

	if not realSpec[prefSpec] then
		if requirePref then
			return false
		end

		doesMatch = ID
	end

	if ignoreSpec and realSpec == ignoreSpec then
		return false
	end

	return doesMatch
end

local function FindInGroup(group, spec)
	for i = 1, getn(group) do
		if currentRaidGroupRolesByIndex[group[i] ][roleIndex[spec] ] then
			return true
		end
	end

	return false
end

local function FindNonNA(unassignedMembers, zonesByIndex)
	for _,v in pairs(unassignedMembers) do
		if zonesByIndex[v] ~= "na" then
			return true
		end
	end
	return false
end

local function DynamicBGArrange(unassignedMembers)
	local realZone = GetRealZoneText()

	if loc.arathi and loc.eots and (realZone ~= loc.arathi and realZone ~= loc.eots) then
		return false
	end

	local optimalSetup = {{},{},{},{},{},{},{},{}}
	local zone = nil

	local zonesByIndex = {}

	for i = 1, getn(unassignedMembers) do
		if UnitIsUnit("raid" .. unassignedMembers[i], "player") then
			zone = GetMinimapZoneText()
		else
			_,_,_,_,_,_,zone = GetRaidRosterInfo(unassignedMembers[i])
		end

		if not zone or zone == "Offline" then
			zone = "na"
		end

		zonesByIndex[unassignedMembers[i] ] = zone
	end

	local wGroup = 1
	while FindNonNA(unassignedMembers, zonesByIndex) do
		local x = 0
		while x < getn(unassignedMembers) do
			x = x + 1
			local cZone = zonesByIndex[unassignedMembers[x] ]
			if cZone ~= "na" then
				while x <= getn(unassignedMembers) do
					if zonesByIndex[unassignedMembers[x] ] == cZone then
						if getn(optimalSetup[wGroup]) > 4 then
							wGroup = wGroup + 1
							if wGroup > 8 then --i don't think this is possible if you're actually in ab/eots so i'm just going to make a really simple failsafe
								for a = 1, 8 do
									if getn(optimalSetup[a]) < 5 then
										wGroup = a
									end
								end
							end
						end

						table.insert(optimalSetup[wGroup], table.remove(unassignedMembers, x))
						x = x - 1
					end

					x = x + 1
				end
			end
		end

		wGroup = wGroup + 1

		if wGroup > 8 then
			for a = 1, 8 do
				if getn(optimalSetup[a]) < 5 then
					wGroup = a
				end
			end
		end
	end

	for a = 1, 8 do
		if getn(optimalSetup[a]) < 5 then
			wGroup = a
		end
	end

	while getn(unassignedMembers) > 0 do
		if getn(optimalSetup[wGroup]) > 4 then
			for a = 1, 8 do
				if getn(optimalSetup[a]) < 5 then
					wGroup = a
				end
			end
		end

		table.insert(optimalSetup[wGroup], table.remove(unassignedMembers))
	end

	return unassignedMembers, optimalSetup
end

local function RemoveIndex(optimalSetup, index)
	for i = 1, 8 do
		local j = 1
		while j <= getn(optimalSetup[i]) do
			local x = optimalSetup[i][j]
			if x and x == index then
				table.remove(optimalSetup[i], j)
				j = j - 1
			end
			j = j + 1
		end
	end

	return optimalSetup
end

local function IgnoreExcludedGroups(optimalSetup, unassignedMembers)
	if ignoredExcluded then
		return optimalSetup, unassignedMembers
	end

	ignoredExcluded = true

	for i = 1, 8 do
		if GrOM_Vars.exclude[i] then
			for k = 1, getn(optimalSetup[i]) do
				table.insert(unassignedMembers,	optimalSetup[i][k])
			end
			optimalSetup[i] = {}
			for j = 1, 40 do
				local n,_,g = GetRaidRosterInfo(j)
				if n and g and g == i then
					optimalSetup = RemoveIndex(optimalSetup, j)
					table.insert(optimalSetup[i], j)
				end
			end
		end
	end

	local b = 1
	while b <= getn(unassignedMembers) do
		local name,_,group = GetRaidRosterInfo(unassignedMembers[b])
		if name and group then
			if GrOM_Vars.exclude[group] then
				table.remove(unassignedMembers, b)
				b = b - 1
			end
		end
		b = b + 1
	end

	return optimalSetup, unassignedMembers
end

local function IsGroupLimitMet(limit, group)  --[1] = class, [2] = ROLE, [3] = number, [4] = enforce, [5] = operator (& or @)
	local count = 0

	for i = 1, getn(group) do
		local matchClass, matchSpec = false, false

		if limit[2] and currentRaidGroupRolesByIndex[group[i] ][roleIndex[limit[2] ] ] then
			matchSpec = true
		end

		if limit[1] and select(2, UnitClass("raid"..group[i])) == limit[1] then
			matchClass = true
		end

		if limit[1] and limit[2] then
			if matchSpec and matchClass then
				count = count + 1
			elseif limit[5] == "@" then
				if matchSpec or matchClass then
					count = count + 1
				end
			end
		elseif limit[1] then
			if matchClass then
				count = count + 1
			end
		else
			if matchSpec then
				count = count + 1
			end
		end
	end

	if count >= limit[3] then
		return true
	end

	return false
end

local function ValidateLimits(...)
	local limit
	local maxN = select("#", ...)
	local limitT = {}

	if maxN < 1 then
		return false
	end

	for i = 1, maxN do
		limit = select(i, ...)
		if type(limit) ~= "string" then
			return false
		end

		local _,_,enforce, number, class, operator, spec = limit:find("^(%+?)(%d+)([^%-]+)([@&])([^%-]+)$")

		if not (class and spec and number) then
			return false
		end

		number = tonumber(number)

		if not number then
			return false
		end

		if class == "*" and spec == "*" then
			return false
		end

		local limTE = {}

		if class == "*" then
			limTE[1] = false
		else
			if not defaultClassRoles[class] then
				return false
			else
				limTE[1] = class
			end
		end

		if spec == "*" then
			limTE[2] = false
		else
			if not roleIndex[spec] then
				return false
			else
				limTE[2] = spec
			end
		end

		if number < 0 or number > 5 then
			return false
		end

		limTE[3] = number
		limTE[4] = (enforce ~= "")
		limTE[5] = operator
		table.insert(limitT, limTE)
	end

	return limitT
end

--/print GrOM.TestValidateLimits("+1SHAMAN@*,+1PALADIN@*,+1PRIEST&MANA_BATTERY")
--/print GrOM.TestValidateLimits("1PRIEST&MANA_BATTERY")
--[[function GrOM.TestValidateLimits(x)
	return ValidateLimits(strsplit(",", x))
end]]

local defaultTemplate = {{
	[1] = "group 1",
	[2] = "limit +1PALADIN@*",
	[3] = "allgroups",
	[4] = {nil, "TANK", true},
	[5] = "nolimit",
	[6] = "fail 16",
	[7] = "group 1",
	[8] = "limit +1PALADIN@*",
	[9] = "allgroups",
	[10] = {nil, "TANK", true},
	[11] = "nolimit",
	[12] = {"DRUID", "TANK_BUFFER", true},
	[13] = {"PALADIN", "TANK_BUFFER"},
	[14] = {"WARLOCK", "TANK_BUFFER"},
	[15] = "end",
	[16] = "limit +1SHAMAN@*,+1PALADIN@*,+1PRIEST&MANA_BATTERY",
	[17] = {nil, "HEALER", true},
	[18] = "fail 24",
	[19] = {nil, "MANA_BATTERY", true},
	[20] = "fail 24",
	[21] = {nil, "HEAL_BUFFER", true},
	[22] = {nil, "HEALER", true, nil, true},
	[23] = "end",
	[24] = {nil, "MELEE_DPS", true},
	[25] = "fail 30",
	[26] = "limit 1*@MELEE_BUFFER",
	[27] = {nil, "MELEE_BUFFER", true},
	[28] = {nil, "MELEE_DPS", true, nil, true},
	[29] = "end",
	[30] = {nil, "RANGE_DPS", true},
	[31] = "fail 36",
	[32] = "limit 1*@RANGE_BUFFER",
	[33] = {nil, "RANGE_BUFFER", true},
	[34] = {nil, "RANGE_DPS", true, nil, true},
	[35] = "end",
	[36] = "limit 1*@SPELL_BUFFER,+1WARLOCK@*",
	[37] = {nil, "SPELL_DPS", true},
	[38] = {nil, "SPELL_BUFFER", true},
	[39] = {"WARLOCK", "SPELL_DPS", false, nil, true},
	[40] = {nil, "SPELL_DPS", true, nil, true}
	},
	{
	[1] = "spec TANK 6",
	[2] = "spec HEALER 11",
	[3] = "spec MELEE_BUFFER 18",
	[4] = "spec RANGE_BUFFER 20",
	[5] = "go 26",
	[6] = "limit +1SHAMAN@*,+1PALADIN@*,+1PRIEST&MANA_BATTERY,+1WARLOCK@*",
	[7] = {nil, "TANK_BUFFER", true, nil, true},
	[8] = {nil, "TANK", true, nil, true},
	[9] = "spec HEALER 11",
	[10] = "go 26",
	[11] = "limit +1SHAMAN@*,+1PALADIN@*,+1PRIEST&MANA_BATTERY",
	[12] = "pass 3",
	[13] = "nolimit",
	[14] = "anypass",
	[15] = {nil, "HEAL_BUFFER", true, nil, true},
	[16] = {nil, "MANA_BATTERY", true, nil, true},
	[17] = "go 26",
	[18] = {nil, "MELEE_DPS", true, nil, true},
	[19] = "go 26",
	[20] = "limit +1SHAMAN@*,+1PALADIN@*,+1PRIEST&MANA_BATTERY,+1WARLOCK@*",
	[21] = "pass 3",
	[22] = "nolimit",
	[23] = "anypass",
	[24] = {nil, "RANGE_DPS", true, nil, true},
	[25] = {nil, "MELEE_DPS", true, nil, true},
	[26] = "limit +1SHAMAN@*,+1PALADIN@*,+1PRIEST&MANA_BATTERY,+1WARLOCK@*",
	[27] = "pass 3",
	[28] = "nolimit",
	[29] = "anypass",
	[30] = {"WARLOCK", "SPELL_DPS"},
	[31] = {nil, "SPELL_DPS", false, nil, true}
	}}

function GrOM.Arrange(optimalSetup, unassignedMembers)
	GrOM.ARRANGE_BREAK = GrOM.ARRANGE_BREAK + 1

	if GrOM.ARRANGE_BREAK and GrOM.ARRANGE_BREAK > 100 then
		dout("Bad user template: Infinite loop terminated. Please select another template or use the default.")
		return false
	end

	local optimalSetup = optimalSetup or {{},{},{},{},{},{},{},{}}

	local unassignedMembers = unassignedMembers or GetAllRaidIDs()

	optimalSetup, unassignedMembers = IgnoreExcludedGroups(optimalSetup, unassignedMembers)

	local x = FindWorkingGroup(optimalSetup, unassignedMembers)

	local currPass = floor(GrOM.ARRANGE_BREAK / GetMaxGroups(unassignedMembers, optimalSetup) )  --this is really ugly

	if x == -2 then
		dout("bad maxgroups")
		return false
	end

	if x == -1 or (not unassignedMembers[1]) then
		return optimalSetup
	end

	local group = optimalSetup[x]

	local match = false

	local class, prefSpec, requirePref, ignoreSpec = nil, nil, false, nil

	local lastMatch, lastIndex = -1, -1

	local usePreviousGroup = false --obsolete?

	local limits = nil  --[1] = class, [2] = ROLE, [3] = number, [4] = enforce, [5] = operator (& or @)

	local minPass = 0

	local function IsLimitedUnit(lid)
		if not limits then
			return false
		end

		local retVal = "."

		for il1 = 1, getn(limits) do
			local l1t = limits[il1]
			if l1t[1] and l1t[2] then
				if (currentRaidGroupRolesByIndex[lid][roleIndex[l1t[2] ] ] and select(2, UnitClass("raid"..lid)) == l1t[1]) or (l1t[5] == "@" and (currentRaidGroupRolesByIndex[lid][roleIndex[l1t[2] ] ] or select(2, UnitClass("raid"..lid)) == l1t[1]) ) then
					if IsGroupLimitMet(l1t, group) then
						if not l1t[4] then
							if retVal ~= "deny" then
								retVal = "permit"
							end
						else
							retVal = "deny"
						end
					end
				end
			elseif l1t[1] then
				if select(2, UnitClass("raid"..lid)) == l1t[1] then
					if IsGroupLimitMet(l1t, group) then
						if not l1t[4] then
							if retVal ~= "deny" then
								retVal = "permit"
							end
						else
							retVal = "deny"
						end
					end
				end
			else
				if currentRaidGroupRolesByIndex[lid][roleIndex[l1t[2] ] ] then
					if IsGroupLimitMet(l1t, group) then
						if not l1t[4] then
							if retVal ~= "deny" then
								retVal = "permit"
							end
						else
							retVal = "deny"
						end
					end
				end
			end
		end

		if retVal == "." then
			return false
		end

		return retVal
	end

	local function HandleMatchup(i, v)
		if i == 1 then
			lastMatch = -1
			match = false
		end

		if not v then
			match = false
			return
		end

		local isMatch = DoesIDMatchClassAndSpec(v, class, prefSpec, requirePref, ignoreSpec)

		if isMatch then
			local isLimitedHow = IsLimitedUnit(v)
			if limits and isLimitedHow then
				if isLimitedHow == "permit" then
					if lastMatch == -1 then
						lastMatch = v
						lastIndex = i
					end
				end
			else
				if isMatch == 41 then
					match = true
					table.remove(unassignedMembers, i)
					if usePreviousGroup then --obsolete?
						table.insert(optimalSetup[x-1], v)
					else
						table.insert(group, v)
					end

					return true
				elseif lastMatch == -1 then
					lastMatch = v
					lastIndex = i
				end
			end
		end

		if i == getn(unassignedMembers) and (not match) and lastMatch > 0 then
			match = true
			table.remove(unassignedMembers, lastIndex)
			if usePreviousGroup then
				table.insert(optimalSetup[x-1], lastMatch)
			else
				table.insert(group, lastMatch)
			end

			return true
		end

		--this shouldn't ever happen
		if i == getn(unassignedMembers) and match then
			match = false
		end
	end

	local function myForeachi(tab, func)
		if getn(tab) == 0 then
			match = false
		end
		for k, v in pairs(tab) do
			if func(k,v) then return end
		end
	end

	local template = {}

	--if a template is selected, and it's valid, use it
	local OKGroups = nil
	if GrOM_Vars.userTemplateToUse then
		if not ( GrOM.userTemplates[GrOM_Vars.userTemplateToUse ] and ValidateTemplate(GrOM.userTemplates[GrOM_Vars.userTemplateToUse ]) ) then
			dout("Bad user template. Please select another template or use the default.")
			return false
		end

		if not group[1] then
			template = GrOM.userTemplates[GrOM_Vars.userTemplateToUse ][2]
		else
			template = GrOM.userTemplates[GrOM_Vars.userTemplateToUse ][3]
		end
		
		if currPass > 2 then
			template = GrOM.userTemplates[GrOM_Vars.userTemplateToUse ][3]
		end
	--otherwise use the default
	else
		if not group[1] then
			template = defaultTemplate[1]
		else
			template = defaultTemplate[2]
		end
		
		if currPass > 2 then
			template = defaultTemplate[2]
		end
	end

	local tc = 0
	while tc < getn(template) and tcount(group) < 5 do
		tc = tc + 1

		if type(template[tc]) == "string" then
			if template[tc] == "allgroups" then
				OKGroups = nil
			elseif template[tc] == "anypass" then
				minPass = 0
			end

			if ((not OKGroups) or myTFind(OKGroups, x)) and (currPass >= minPass) then
				local cmd, sc = GetCmd(template[tc])
				if cmd and sc then
					if cmd == "fail" and not match then
						sc = tonumber(sc)
						if sc and sc > tc then
							tc = sc - 1
						else
							dout("Bad user template. Please select another template or use the default. Aborting. Error at index " .. tc .. ". Usage: fail # (# must be > the current index)")
							return false
						end
					elseif cmd == "dynbg" then
						unassignedMembers, optimalSetup = DynamicBGArrange(unassignedMembers)
						if not optimalSetup then
							cout(loc.badlocationerr)
							return false
						else
							return optimalSetup
						end
					elseif cmd == "go" then
						sc = tonumber(sc)
						if sc and sc > tc then
							tc = sc - 1
						else
							dout("Bad user template. Please select another template or use the default. Aborting. Error at index " .. tc .. ". Usage: go # (# must be > the current index)")
							return false
						end
					elseif cmd == "pass" then
						sc = tonumber(sc)
						if sc and sc > 0 then
							minPass = sc
						else
							dout("Bad user template. Please select another template or use the default. Aborting. Error at index " .. tc .. ". Usage: pass # (# must be > 0")
							return false
						end
					elseif cmd == "limit" then
						limits = ValidateLimits(strsplit(",", sc))
						if not limits then
							dout("Bad user template. Please select another template or use the default. Aborting. Error at index " .. tc .. ". Usage: limit \<limitations\>\nSee templates.lua for limitations syntax.")
							return false
						end
					elseif cmd == "nolimit" then
						limits = nil
					elseif cmd == "group" then
						OKGroups = ValidateGroups(strsplit(",", sc))
						if not OKGroups then
							dout("Bad user template. Please select another template or use the default. Aborting. Error at index " .. tc .. ". Usage: group #\[,#\[,#\[,#\[,#\[,#\[,#\[,#\]\]\]\]\]\]\]")
							return false
						end
					elseif cmd == "spec" then
						local _,_,n,g = sc:find("(.+) (%d+)")

						if g then
							g = tonumber(g)
						end

						if g and g > tc and n and roleIndex[n] then
							if FindInGroup(group, n) then
								tc = g - 1
							end
						else
							dout("Bad user template. Please select another template or use the default. Aborting. Error at index " .. tc .. ". Usage: spec ROLE # (# must be > the current index)")
							return false
						end
					elseif cmd == "end" then
						tc = getn(template)
					end
				end
			end
		elseif type(template[tc]) == "table" then
			if not ValidateTemplateTable(template[tc]) then
				dout("Bad user template. Please select another template or use the default. Aborting. Error at index " .. tc .. ". Usage: {class\|nil, prefSpec, requirePref?\[, ignoreSpec\|nil\[, loop?\]\]}")
				return false
			end
			if ((not OKGroups) or myTFind(OKGroups, x)) and (currPass >= minPass) then
				class, prefSpec, requirePref, ignoreSpec, loopWhile = template[tc][1], template[tc][2], template[tc][3], template[tc][4], template[tc][5]
				if prefSpec then
					prefSpec = roleIndex[prefSpec]
				end
				myForeachi(unassignedMembers, HandleMatchup)
				while loopWhile and tcount(group) < 5 and match do
					myForeachi(unassignedMembers, HandleMatchup)
				end
			end
		end
	end

	optimalSetup[x] = group
	return GrOM.Arrange(optimalSetup, unassignedMembers)

	--[[
	--no template, use defaults
	if not group[1] then
		--we're creating a new group
		--if there is a tank available, add it and
			--if there is another tank available, add it too
			--if there is a tree druid, add it
			--if there is a pally that is not a tank and is a "tank buffer" available, add it
			--otherwise if there is a pally that isn't a tank available, add it
			--if there is a warlock that is a "tank buffer" available, add it, otherwise, if there is any other warlock available, add it
			--stop
		--else if there is a healer available, add it and
			--if there is a mana battery, add it
			--if there is any free "heal buffer" add it
			--add as many other healers as there are free
			--stop
		--else if there is a melee dps available, add it and
			--if there is any melee buffer available, add it
			--find as many melee dps as are available, and add them
			--stop
		--else do the same as above for ranged and then spell dps


		class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["TANK"], true, nil
		myForeachi(unassignedMembers, HandleMatchup)
		if match then
			class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["TANK"], true, nil
			myForeachi(unassignedMembers, HandleMatchup)
			--tree druids ... i dunno why i thought i needed to do it this way >.<
			if not match then
				class, prefSpec, requirePref, ignoreSpec = "DRUID", roleIndex["TANK_BUFFER"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)
			elseif x > 1 then
				usePreviousGroup = true
				class, prefSpec, requirePref, ignoreSpec = "DRUID", roleIndex["TANK_BUFFER"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)
				if match and getn(optimalSetup[x-1]) == 6 then
					for gc=1, 5 do
						if group[i] and currentRaidGroupRolesByIndex[group[i] ][roleIndex["TANK"] ] then
							table.insert(unassignedMembers,table.remove(group, i))
							break
						end
					end
					for gp=1, 6 do
						if optimalSetup[x-1][i] and currentRaidGroupRolesByIndex[optimalSetup[x-1][i] ][roleIndex["TANK"] ] then
							table.insert(group, table.remove(optimalSetup[x-1], i))
							break
						end
					end
				end
				usePreviousGroup = false
			end
			--end tree druids

			class, prefSpec, requirePref, ignoreSpec = "DRUID", roleIndex["TANK_BUFFER"], true, roleIndex["TANK"]
			myForeachi(unassignedMembers, HandleMatchup)

			class, prefSpec, requirePref, ignoreSpec = "PALADIN", roleIndex["TANK_BUFFER"], false, roleIndex["TANK"]
			myForeachi(unassignedMembers, HandleMatchup)

			class, prefSpec, requirePref, ignoreSpec = "WARLOCK", roleIndex["TANK_BUFFER"], false, nil
			myForeachi(unassignedMembers, HandleMatchup)
		else
			class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["HEALER"], true, nil
			myForeachi(unassignedMembers, HandleMatchup)

			if match then
				class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["MANA_BATTERY"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)

				class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["HEAL_BUFFER"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)

				match = true
				while tcount(group) < 5 and match do
					class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["HEALER"], true, nil
					myForeachi(unassignedMembers, HandleMatchup)
				end
			else
				class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["MELEE_DPS"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)

				if match then
					class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["MELEE_BUFFER"], true, nil
					myForeachi(unassignedMembers, HandleMatchup)

					match = true
					while tcount(group) < 5 and match do
						class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["MELEE_DPS"], true, nil
						myForeachi(unassignedMembers, HandleMatchup)
					end
				else
					class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["RANGE_DPS"], true, nil
					myForeachi(unassignedMembers, HandleMatchup)

					if match then
						class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["RANGE_BUFFER"], true, nil
						myForeachi(unassignedMembers, HandleMatchup)

						match = true
						while tcount(group) < 5 and match do
							class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["RANGE_DPS"], true, nil
							myForeachi(unassignedMembers, HandleMatchup)
						end
					else
						class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["SPELL_DPS"], true, nil
						myForeachi(unassignedMembers, HandleMatchup)

						if match then
							class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["SPELL_BUFFER"], true, nil
							myForeachi(unassignedMembers, HandleMatchup)

							match = true
							while tcount(group) < 5 and match do
								class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["SPELL_DPS"], true, nil
								myForeachi(unassignedMembers, HandleMatchup)
							end
						else
							cout(loc.arrangefailederr)
							return false
						end
					end
				end
			end
		end
	else
		--we're filling in empty space
		--if there is a healer anywhere in the group
			--find any available heal buffers or mana batteries and add them
		--if there is a tank anywhere in the group
			--find any available tanks or tank buffers and add them
		--if there are any melee buffers in the group, add any available melee dps
		--if there are any ranged buffers in the group, add any available ranged dps and then melee dps
		--if there are any spell buffers in the group, add any available spell dps
		if FindInGroup(group, "HEALER") then
			match = true
			while tcount(group) < 5 and match do
				class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["HEAL_BUFFER"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)
			end

			match = true
			while tcount(group) < 5 and match do
				class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["MANA_BATTERY"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)
			end
		end
		if FindInGroup(group, "TANK") then
			match = true
			while tcount(group) < 5 and match do
				class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["TANK_BUFFER"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)
			end

			match = true
			while tcount(group) < 5 and match do
				class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["TANK"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)
			end
		end
		if FindInGroup(group, "MELEE_BUFFER") then
			match = true
			while tcount(group) < 5 and match do
				class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["MELEE_DPS"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)
			end
		end
		if FindInGroup(group, "RANGE_BUFFER") then
			match = true
			while tcount(group) < 5 and match do
				class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["RANGE_DPS"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)
			end

			match = true
			while tcount(group) < 5 and match do
				class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["MELEE_DPS"], true, nil
				myForeachi(unassignedMembers, HandleMatchup)
			end
		end
		--fill in any remaining spots with whatever is available
		match = true
		while tcount(group) < 5 and match do
			class, prefSpec, requirePref, ignoreSpec = nil, roleIndex["MELEE_DPS"], false, nil
			myForeachi(unassignedMembers, HandleMatchup)
		end
	end
	optimalSetup[x] = group

	return GrOM.Arrange(optimalSetup, unassignedMembers)]]
end

local function UMfind(t, n)
	for i = 1, getn(t) do
		for j = 1, getn(t[i]) do
			if t[i][j] == n then
				return true
			end
		end
	end
end

--return the index in unassignedMembers that corrisponds to a member with the same roles as find, which is not in restoreState
local function FindSimilarExtantMember(restoreState, unassignedMembers, find)
	if not find then
		return false
	end

	for i = 1, getn(unassignedMembers) do
		if not UMfind(restoreState, GetPVEorPVPUName("raid"..unassignedMembers[i]) ) then
			local matchS, a = true, currentRaidGroupRolesByIndex[unassignedMembers[i] ]
			for j = 1, getn(a) do		
				if a[j] ~= find[j] then
					matchS = false
					break
				end
			end
			
			if matchS then
				return i
			end
		end
	end
	
	return false
end


local lastTemplate = nil

function GrOM.RestoreRaid(restoreState)
	if GrOM_Vars.userTemplateToUse == "AB_EOTS_AUTO_AKRYN" then
		lastTemplate = GrOM_Vars.userTemplateToUse
		GrOM_Vars.userTemplateToUse = nil
	end

	GrOM.ARRANGE_BREAK = 0
	
	local _,_,rModeA, rModeB = GrOM.GetCurrentRestoreMethod()

	rModeA = (rModeA == "smart")
	rModeB = (rModeB == "separate")
	
	resetErrorTimer = 0
	groupFullErr = false
	optimalRaidGroupPlacement = {}

	BuildRolesByIndexTable()
	BuildIndexByNameTable()

	local myRoles = GetMyRoles()
	if not myRoles then
		cout(loc.talentserr)
		myRoles = defaultClassRoles[select(2,UnitClass("player"))]
	end
	currentRaidGroupRolesByIndex[currentRaidGroupIndexByName[UnitName("player")] ] = myRoles

	FillInEmptyRoles(1)

	local optimalSetup, unassignedMembers = {{},{},{},{},{},{},{},{}}, GetAllRaidIDs()

	local i, j, k = 0, 0, 0

	if rModeA then
		for rmac = 1, 8 do
			local rmav = restoreState[rmac]
			if rmav then
				for rmacb = 1, getn(rmav) do
					if not currentRaidGroupIndexByName[rmav[rmacb] ] then  --if this person isn't in the raid
						local rmanew = FindSimilarExtantMember(restoreState, unassignedMembers, GrOM_Vars.RolesByNameCache[rmav[rmacb] ])
						if rmanew then
							table.insert(optimalSetup[rmac], table.remove(unassignedMembers, rmanew))
						end
					end
				end
			end
		end
	end	
	
	while i < getn(unassignedMembers) do
		i = i + 1
		j=0
		while j < 8 do
			j = j + 1
			k=0
			local z = restoreState[j]
			if not z then
				break
			end
			while k < getn(z) do
				k = k + 1

				local y = z[k]
				if y and currentRaidGroupIndexByName[y] and (unassignedMembers[i] == currentRaidGroupIndexByName[y]) then
					table.insert(optimalSetup[j], table.remove(unassignedMembers, i))
					i = i - 1
					j, k = 9, 6
				end
			end
		end
	end
	
	if rModeB and getn(unassignedMembers) > 0 then
		i = 0
		while i < getn(unassignedMembers) do
			i = i + 1
			for j=8, 1, -1 do
				if getn(optimalSetup[j]) < 5 then
					table.insert(optimalSetup[j], table.remove(unassignedMembers, i))
					i = i - 1
					break
				end
			end
		end
	end

	ignoredExcluded = false
	optimalSetup = GrOM.Arrange(optimalSetup, unassignedMembers)

	if not optimalSetup then
		return
	end

	local g = 1

	local group = optimalSetup[g]

	while group and group ~= {} do
		for i=1, 5 do
			if group[i] then
				optimalRaidGroupPlacement[group[i]] = g
			end
		end
		g = g + 1
		group = optimalSetup[g]
	end

	if not GrOM_Vars.AlwaysSure then
		StaticPopup_Show("GOMatic_DO_ARRANGE")
	else
		GrOM.CreateArrangementFrame()
	end
end

function GrOM.SaveRaid(name)
	GrOM.SaveName = nil

	local raid = {{},{},{},{},{},{},{},{}}

	for i=1, 40 do
		local n, _, g = GetRaidRosterInfo(i)
		if n and not (GrOM_Vars.excludeSaves and GrOM_Vars.exclude[g]) then
			table.insert(raid[g], n)
		end
	end

	GrOM_Vars.SavedRaidLayouts[name] = raid

	GrOM.UpdateButtons()
	GrOM.UpdateSavesPane()

	cout(string.format(loc.raidsaved, name))
end

function GrOM.DeleteAllSavedRaids()
	GrOM_Vars.SavedRaidLayouts = {}

	GrOM.UpdateButtons()
	GrOM.UpdateSavesPane()

	cout(loc.deletedall)
end

local function CheckForMove()
	local x, y = lastRaidSetup, currentRaidGroupPlacement

	if getn(x) ~= getn(y) then
		return true
	end

	for i = 1, getn(y) do
		if not x[i] or x[i] ~= y[i] then
			return true
		end
	end

	return false
end

local function GrOM_SetRaidSubgroup(i, g)
	table.insert(movementQueue, {"set", i, g})
	table.insert(lockedIDs, i)
	currentRaidGroupPlacement[i] = g --hopefully we don't need to check if this group is full here
	queuedSetsByGroup[g] = queuedSetsByGroup[g] - 1
	if queuedSetsByGroup[g] < 1 then
		table.insert(lockedGroups, g)
	end
end

local function GrOM_SwapRaidSubgroup(i1, i2)
	table.insert(movementQueue, {"swap", i1, i2})
	table.insert(lockedIDs, i1)
	table.insert(lockedIDs, i2)
	local x = currentRaidGroupPlacement[i1]
	currentRaidGroupPlacement[i1] = currentRaidGroupPlacement[i2]
	currentRaidGroupPlacement[i2] = x
end

local function ExecuteQueuedMoves()
	lockedIDs = {}
	lockedGroups = {}
	queuedSetsByGroup = {5,5,5,5,5,5,5,5}
	BuildCurrentPlacementTable()

	if getn(movementQueue) < 1 then
		return false
	end

	while getn(movementQueue) > 0 do
		local move = table.remove(movementQueue)
		if move[1] == "set" then
			SetRaidSubgroup(move[2], move[3])
		else
			SwapRaidSubgroup(move[2], move[3])
		end
	end

	return true
end

local function BuildQueuedSetsTable() --this table represents the number of times we can queue a SetRaidSubgroup call for each group
	queuedSetsByGroup = {5,5,5,5,5,5,5,5}

	for i = 1, 40 do
		if currentRaidGroupPlacement[i] then
			local x = currentRaidGroupPlacement[i]

			if x > 0 and x < 9 then
				queuedSetsByGroup[x] = queuedSetsByGroup[x] - 1

				if queuedSetsByGroup[x] < 1 then
					table.insert(lockedGroups, x)
				end
			end
		end
	end
end

local function DoArrangement()
	if InCombatLockdown() then
		GrOM.Cancel()
		cout(loc.combaterr)
	end

	if resetErrorTimer + GrOM.resetErrorConditionThrottle < GetTime() then
		resetErrorTimer = GetTime()
		groupFullErr = false
		lastRaidSetup = {}
		movementQueue = {}
		lockedIDs = {}
		lockedGroups = {}
		queuedSetsByGroup = {5,5,5,5,5,5,5,5}
	end

	if GrOM.ArrangementFrameBreakCounter > GrOM.ArrangementFrameBreakMaximum then
		if not groupFullErr then
			StaticPopup_Show("GOMatic_ERROR_TOOMANYFAILEDMOVES")
		end
		delayedActionInProgress = false
		GrOM.UpdateButtons()
		GOMatic_ArrangementFrame:SetScript("OnUpdate", nil)
		GOMatic_ArrangementFrame:SetScript("OnEvent", nil)
		GOMatic_ArrangementFrame = nil
	end

	GOMatic_ArrangementFrameLastUpdate = GetTime()

	if not movementQueue[1] then
		if BuildCurrentPlacementTable() then
			return false
		end

		BuildQueuedSetsTable()
	end

	RepairOptimalPlacementTable()

	if not movementQueue[1] then
		if not CheckForMove() then
			return false
		else
			lastRaidSetup = CopyTable(currentRaidGroupPlacement)
		end
	end

	if getn(movementQueue) >= GrOM_Vars.autoThrottle then
		return ExecuteQueuedMoves()
	end

	local emptySlot, raidIDToMove = FindEmptySlot()

	if raidIDToMove and (emptySlot > 0) and (emptySlot < 9) and (not groupFullErr) then
		GrOM_SetRaidSubgroup(raidIDToMove, emptySlot)
		return DoArrangement()
	end

	local raidIDToMove, currentGroup, destGroup = FindUnmovedID()

	if not raidIDToMove then
		return ExecuteQueuedMoves()
	end

	local raidIDToMoveB = FindIDToMoveToGroup(currentGroup, destGroup)

	if raidIDToMoveB then
		GrOM_SwapRaidSubgroup(raidIDToMove, raidIDToMoveB)
	else
		raidIDToMoveB = FindUnmovedID(currentGroup)
		if not raidIDToMoveB then
			return ExecuteQueuedMoves()
		end

		GrOM_SwapRaidSubgroup(raidIDToMove, raidIDToMoveB)
	end

	return DoArrangement()
end

function GrOM.RaidIsArranged()
	local isArranged = true

	if FindUnmovedID() then
		isArranged = false
	end

	if isArranged then
		cout(loc.arrangedone)
	end

	return isArranged
end

function GOMatic_ArrangementFrame_OnEvent(self, event, arg1)
	if arg1:find(ERR_RAID_MEMBER_REMOVED_S:format("(.+)")) or arg1:match(ERR_RAID_YOU_LEFT) then
		StaticPopup_Show("GOMatic_AUTO_ABORT")
		delayedActionInProgress = false
		GrOM.UpdateButtons()
		self:SetScript("OnUpdate", nil)
		self:UnregisterEvent("CHAT_MSG_SYSTEM")
		self:SetScript("OnEvent", nil)
		GOMatic_ArrangementFrame = nil
	end
end

function GOMatic_ArrangementFrame_OnUpdate(self)
	if (GrOM.ArrangementFrameUpdateThrottle + GOMatic_ArrangementFrameLastUpdate < GetTime()) then
		if GrOM.RaidIsArranged() then
			if lastTemplate then
				GrOM_Vars.userTemplateToUse = lastTemplate
				lastTemplate = nil
			end

			delayedActionInProgress = false
			GrOM.UpdateButtons()
			self:SetScript("OnUpdate", nil)
			self:UnregisterEvent("CHAT_MSG_SYSTEM")
			self:SetScript("OnEvent", nil)
			GOMatic_ArrangementFrame = nil
			return
		end

		if not (IsRaidLeader() or IsRaidOfficer()) then
			delayedActionInProgress = false
			GrOM.UpdateButtons()
			self:SetScript("OnUpdate", nil)
			self:UnregisterEvent("CHAT_MSG_SYSTEM")
			self:SetScript("OnEvent", nil)
			GOMatic_ArrangementFrame = nil
			return
		end

		lockedIDs = {}
		lockedGroups = {}
		movementQueue = {}
		queuedSetsByGroup = {5,5,5,5,5,5,5,5}

		local x = (DoArrangement()) and 0 or 1
		GrOM.ArrangementFrameBreakCounter = GrOM.ArrangementFrameBreakCounter + x
	end
end

GrOM.RestoreVsSave = true

function GrOM.ShowRestoreMenu()
	if delayedActionInProgress then
		cout(loc.pleasewaiterr)
		return
	end

	GrOM.RestoreVsSave = true
	ToggleDropDownMenu(1, nil, GOMaticDropDown, "GOMaticRestore", 70, 0)
end

function GrOM.ShowSaveMenu()
	GrOM.RestoreVsSave = false
	ToggleDropDownMenu(1, nil, GOMaticDropDown, "GOMaticSave", 70, 0)
end

function GrOM.AddSaveOrRestoreMenuItem(k, v)
	local info = {}

	info.text = k

	if not GrOM.RestoreVsSave then
		info.func = GOMaticDropdown_DoSaveMenuItem
		info.arg1 = nil
		info.arg2 = k
	else
		info.func = GOMaticDropdown_DoRestoreMenuItem
		info.arg1 = k
	end

	UIDropDownMenu_AddButton(info)
end

function GrOM.GetCurrentTemplate()
	if not GrOM_Vars.userTemplateToUse then
		return loc.defaultname, loc.defaultname
	end

	local name = GrOM.userTemplates[GrOM_Vars.userTemplateToUse]

	if name and type(name) == "table" then
		name = name[1]
		local full = name

		if name and type(name) == "string" then
			if name:len() > 14 then
				name = name:sub(1, 15) .. "..."
			end

			return full, name
		end
	end

	GrOM_Vars.userTemplateToUse = nil
	return GrOM.GetCurrentTemplate()
end

local currentExtras = "none"

function GrOM.GetCurrentExtrasPane()
	local name, shortname, default = nil, nil, nil

	if currentExtras == "none" then
		name = loc.extrasmenuitem
		default = true
	elseif currentExtras == "exclude" then
		name = loc.excludemenuitem
	elseif currentExtras == "throttle" then
		name = loc.throttlemenuitem
	elseif currentExtras == "MVPs" then
		name = loc.mvpsmenuitem
	elseif currentExtras == "saves" then
		name = loc.savesmenuitem
	end

	if not name then return "???","???",false end

	if name:len() > 13 then
		shortname = name:sub(1, 14) .. "..."
	else
		shortname = name
	end

	return name, shortname, default
end

function GOMaticDropdown_DoTemplateMenuItem(key)
	local template = GrOM.userTemplates[key]

	if template and ValidateTemplate(template) then
		GrOM_Vars.userTemplateToUse = key
	else
		GrOM_Vars.userTemplateToUse = nil
	end

	local x, y = GrOM.GetCurrentTemplate()

	UIDropDownMenu_SetSelectedValue(GOMaticTemplateMenu, x)
	GOMaticTemplateMenuText:SetText(y)
end

function GrOM.AddTemplateMenuItem(key, template)
	local info = {}

	local n = "???"

	if type(template) == "table" then
		if template[1] and type(template[1]) == "string" then
			n = template[1]
		end
	end

	info.text = n
	info.func = GOMaticDropdown_DoTemplateMenuItem
	info.arg1 = key
	UIDropDownMenu_AddButton(info)
end

function GOMaticRestoreMethodDropdown_Initialize()
	local info = {}

	--Dumb - Merge (default)
	info.text = loc.restoremethodpattern:format(loc["dumb"],loc["merge"]) .. loc.defaultmethod
	info.func = GrOM.SetRestoreMode
	info.arg1 = "dumb"
	info.arg2 = "merge"
	UIDropDownMenu_AddButton(info)
	
	--Dumb - Separate
	info = {}
	info.text = loc.restoremethodpattern:format(loc["dumb"],loc["separate"])
	info.func = GrOM.SetRestoreMode
	info.arg1 = "dumb"
	info.arg2 = "separate"
	UIDropDownMenu_AddButton(info)
	
	--Smart - Merge
	info = {}
	info.text = loc.restoremethodpattern:format(loc["smart"],loc["merge"])
	info.func = GrOM.SetRestoreMode
	info.arg1 = "smart"
	info.arg2 = "merge"
	UIDropDownMenu_AddButton(info)
	
	--Smart - Separate
	info = {}
	info.text = loc.restoremethodpattern:format(loc["smart"],loc["separate"])
	info.func = GrOM.SetRestoreMode
	info.arg1 = "smart"
	info.arg2 = "separate"
	UIDropDownMenu_AddButton(info)
end

function GrOM.SetRestoreMode(mode1, mode2)
	GrOM_Vars.restoreMethod = mode1.."-"..mode2
	UIDropDownMenu_SetSelectedValue(GOMaticRestoreMethodMenu, GrOM.GetCurrentRestoreMethod())
	GOMaticRestoreMethodMenuText:SetText(select(2,GrOM.GetCurrentRestoreMethod()))	
end

function GrOM.GetCurrentRestoreMethod()
	local a, b = loc[GrOM_Vars.restoreMethod:match("^([^%-]+)-")], loc[GrOM_Vars.restoreMethod:match("-([^%-]+)")]
	local name = loc.restoremethodpattern:format(a,b)
	
	local short = name
	if name:len() > 15 then
		short = name:sub(1, 15) .. "..."
	end
	
	if GrOM_Vars.restoreMethod == "dumb-merge" then
		name = name .. loc.defaultmethod
	end
	
	return name, short, GrOM_Vars.restoreMethod:match("^([^%-]+)-"), GrOM_Vars.restoreMethod:match("-([^%-]+)")
end

function GOMaticExtrasDropdown_Initialize()
	local info = {}

	--nothing/default
	if select(3, GrOM.GetCurrentExtrasPane()) then
		info.text = loc.extrasmenuitem
	else
		info.text = loc.closemenuitem
	end
	info.func = GrOM.CloseExtrasPane
	info.arg1 = true
	UIDropDownMenu_AddButton(info)

	--blank
	info = {}
	info.text = "   "
	info.isTitle = 1
	UIDropDownMenu_AddButton(info)

	--excludes
	info = {}
	info.text = loc.excludemenuitem
	info.func = GrOM.OpenExtrasPane
	info.arg1 = "exclude"
	UIDropDownMenu_AddButton(info)

	--arrangement throttle
	info = {}
	info.text = loc.throttlemenuitem
	info.func = GrOM.OpenExtrasPane
	info.arg1 = "throttle"
	UIDropDownMenu_AddButton(info)

	--MVPs
	info = {}
	info.text = loc.mvpsmenuitem
	info.func = GrOM.OpenExtrasPane
	info.arg1 = "MVPs"
	UIDropDownMenu_AddButton(info)

	--Saves
	info = {}
	info.text = loc.savesmenuitem
	info.func = GrOM.OpenExtrasPane
	info.arg1 = "saves"
	UIDropDownMenu_AddButton(info)
end

function GOMaticTemplateDropdown_Initialize()
	local info = {}

	info.text = loc.defaultname
	info.func = GOMaticDropdown_DoTemplateMenuItem
	info.arg1 = nil
	UIDropDownMenu_AddButton(info)

	foreach(GrOM.userTemplates, GrOM.AddTemplateMenuItem)
end

function GOMaticDropdown_Initialize()
	local info = {}

	if not GrOM.RestoreVsSave then
		info.text = loc.newmenuitem
		info.func = GOMaticDropdown_DoSaveMenuItem
		info.arg1 = true
		UIDropDownMenu_AddButton(info)

		info = {}
		info.text = "   "
		info.isTitle = 1
		UIDropDownMenu_AddButton(info)
	end

	foreach(GrOM_Vars.SavedRaidLayouts, GrOM.AddSaveOrRestoreMenuItem)

	if GrOM.RestoreVsSave then
		info = {}
		info.text = "   "
		info.isTitle = 1
		UIDropDownMenu_AddButton(info)

		info = {}
		info.text = loc.deleteallmenuitem
		info.func = StaticPopup_Show
		info.arg1 = "GOMatic_DELETE_ALL_SAVES"
		UIDropDownMenu_AddButton(info)
	end
end

function GOMaticDropdown_DoSaveMenuItem(isNew, sc)
	if isNew then
		StaticPopup_Show("GOMatic_NEW_SAVE")
		return
	end
	if sc then
		if GrOM_Vars.SavedRaidLayouts[sc] then
			GrOM.SaveName = sc
			StaticPopup_Show("GOMatic_DO_SAVE", sc)
		else
			GrOM.SaveRaid(sc)
		end
	end
end

function GOMaticDropdown_DoRestoreMenuItem(sc)
	if sc then
		local n = GrOM_Vars.SavedRaidLayouts[sc]
		if n then
			lastAction = {true, sc}
			GrOM.RestoreRaid(n)
		end
	end
end

function GrOM.UpdateButtons(combat)
	CloseMenus()

	local x = InCombatLockdown()
	if combat then
		if combat == 1 then
			x = true
		elseif combat == 0 then
			x = false
		end
	end

	if x then
		if delayedActionInProgress then
			GOMaticAuto:Disable()
			GOMaticPing:Disable()
			GOMaticRestore:Disable()
			GOMaticCancel:Disable()
		else
			GOMaticAuto:Disable()
			GOMaticPing:Enable()
			GOMaticRestore:Disable()
			GOMaticCancel:Disable()
		end
	else
		if delayedActionInProgress then
			GOMaticAuto:Disable()
			GOMaticPing:Disable()
			GOMaticRestore:Disable()
			GOMaticCancel:Enable()
		else
			GOMaticAuto:Enable()
			GOMaticPing:Enable()

			if tcount(GrOM_Vars.SavedRaidLayouts) > 0 then
				GOMaticRestore:Enable()
			else
				GOMaticRestore:Disable()
			end

			GOMaticCancel:Disable()
		end
	end
end

function GrOM.UpdateAutoThrottle(value)
	GrOM_Vars.autoThrottle = value
	GOMaticThrottleText:SetText(loc.throttletext:format(GrOM_Vars.autoThrottle))
end

local checkBoxes = {}

local function CreateExcludeCheck(num)
	if checkBoxes["box" .. num] then
		return false
	end

	local f = CreateFrame("CheckButton", "GOMaticExcludeCheck" .. num, GOMatic, "OptionsCheckButtonTemplate")
	f:SetScript("OnClick", function(self) GrOM_Vars.exclude[num] = self:GetChecked() end)
	f:SetScript("OnShow", function(self) self:SetChecked(GrOM_Vars.exclude[num]) end)
	local textField = _G[f:GetName() .. "Text"]
	--textField:ClearAllPoints()
	--textField:SetPoint("BOTTOMLEFT", f, "TOPLEFT")
	textField:SetText(loc.grouplabel .. num)
	f:SetChecked(GrOM_Vars.exclude[num])
	f:SetHitRectInsets(0, -40, 0, 0)

	return f
end

local function ExcludeHide(resize)
	for i = 1, 8 do
		checkBoxes["box" .. i]:Hide()
	end

	GOMaticExcludeText:Hide()
	GOMaticExcludeSaves:Hide()

	if resize then
		GOMatic:ClearAllPoints()
		local x, y = GOMatic:GetLeft(), GOMatic:GetBottom()
		GOMatic:SetHeight(166)
		GOMatic:SetPoint("BOTTOMLEFT", "WorldFrame", "BOTTOMLEFT", x, y + 90)
	end

	return 90
end

local function ExcludeShow(yOffset)
	checkBoxes["box1"] = CreateExcludeCheck(1) or checkBoxes["box1"]
	checkBoxes["box1"]:Show()
	checkBoxes["box1"]:ClearAllPoints()
	checkBoxes["box1"]:SetPoint("TOPLEFT", "GOMaticHide", "BOTTOMLEFT", 0, -30)

	for i = 2, 8 do
		checkBoxes["box" .. i] = CreateExcludeCheck(i) or checkBoxes["box" .. i]
		checkBoxes["box" .. i]:Show()
		checkBoxes["box" .. i]:ClearAllPoints()
		if i == 5 then
			checkBoxes["box5"]:SetPoint("TOP", checkBoxes["box1"], "BOTTOM", 0, -3)
		else
			checkBoxes["box" .. i]:SetPoint("LEFT", checkBoxes["box" .. i - 1], "RIGHT", 65, 0)
		end
	end

	GOMaticExcludeText:Show()
	GOMaticExcludeSaves:Show()

	GOMatic:ClearAllPoints()
	local x, y = GOMatic:GetLeft(), GOMatic:GetBottom()
	GOMatic:SetHeight(256)
	GOMatic:SetPoint("BOTTOMLEFT", "WorldFrame", "BOTTOMLEFT", x, y - 90 + yOffset)
end

local function ThrottleShow(yOffset)
	GOMaticThrottleSlider:Show()
	GOMaticThrottleText:Show()

	GOMatic:ClearAllPoints()
	local x, y = GOMatic:GetLeft(), GOMatic:GetBottom()
	GOMatic:SetHeight(226)
	GOMatic:SetPoint("BOTTOMLEFT", "WorldFrame", "BOTTOMLEFT", x, y - 60 + yOffset)
end

local function ThrottleHide(resize)
	GOMaticThrottleSlider:Hide()
	GOMaticThrottleText:Hide()

	if resize then
		GOMatic:ClearAllPoints()
		local x, y = GOMatic:GetLeft(), GOMatic:GetBottom()
		GOMatic:SetHeight(166)
		GOMatic:SetPoint("BOTTOMLEFT", "WorldFrame", "BOTTOMLEFT", x, y + 60)
	end

	return 60
end

local function SavesShow(yOffset)
	local h = 220

	GOMaticSavesScrollFrame:Show()
	GOMaticDeleteSave:Show()
	GOMaticRenameSave:Show()
	GOMaticRestoreMethodMenu:Show()
	GOMaticRestoreMethodText:Show()

	GOMatic:ClearAllPoints()
	local x, y = GOMatic:GetLeft(), GOMatic:GetBottom()
	GOMatic:SetHeight(166+h)
	GOMatic:SetPoint("BOTTOMLEFT", "WorldFrame", "BOTTOMLEFT", x, y - h + yOffset)
end

local function SavesHide(resize)
	local h = 220

	GOMaticSavesScrollFrame:Hide()
	GOMaticDeleteSave:Hide()
	GOMaticRenameSave:Hide()
	GOMaticRestoreMethodMenu:Hide()
	GOMaticRestoreMethodText:Hide()

	if resize then
		GOMatic:ClearAllPoints()
		local x, y = GOMatic:GetLeft(), GOMatic:GetBottom()
		GOMatic:SetHeight(166)
		GOMatic:SetPoint("BOTTOMLEFT", "WorldFrame", "BOTTOMLEFT", x, y + h)
	end

	return h
end

local function MVPsShow(yOffset)
	local h = 220

	GOMaticMVPScrollFrame:Show()
	GOMaticAddMVP:Show()
	GOMaticDeleteMVP:Show()
	GOMaticDeleteAllMVPs:Show()
	GOMaticDemoteMVP:Show()
	GOMaticPromoteMVP:Show()

	GOMatic:ClearAllPoints()
	local x, y = GOMatic:GetLeft(), GOMatic:GetBottom()
	GOMatic:SetHeight(166+h)
	GOMatic:SetPoint("BOTTOMLEFT", "WorldFrame", "BOTTOMLEFT", x, y - h + yOffset)
end

local function MVPsHide(resize)
	local h = 220

	GOMaticMVPScrollFrame:Hide()
	GOMaticAddMVP:Hide()
	GOMaticDeleteMVP:Hide()
	GOMaticDeleteAllMVPs:Hide()
	GOMaticDemoteMVP:Hide()
	GOMaticPromoteMVP:Hide()

	if resize then
		GOMatic:ClearAllPoints()
		local x, y = GOMatic:GetLeft(), GOMatic:GetBottom()
		GOMatic:SetHeight(166)
		GOMatic:SetPoint("BOTTOMLEFT", "WorldFrame", "BOTTOMLEFT", x, y + h)
	end

	return h
end

function GrOM.CloseExtrasPane(updatePos)
	local yOffset = 0

	if currentExtras ~= "none" then
		if currentExtras == "exclude" then
			yOffset = ExcludeHide(updatePos)
		elseif currentExtras == "throttle" then
			yOffset = ThrottleHide(updatePos)
		elseif currentExtras == "MVPs" then
			yOffset = MVPsHide(updatePos)
		elseif currentExtras == "saves" then
			yOffset = SavesHide(updatePos)
		end
	end

	currentExtras = "none"

	local x, y = GrOM.GetCurrentExtrasPane()

	UIDropDownMenu_SetSelectedValue(GOMaticExtrasMenu, x)
	GOMaticExtrasMenuText:SetText(y)

	return yOffset
end

local excludeShowing = false --obsolete?

--obsolete?
function GrOM.ExcludeClick(hideOnly)
	if hideOnly and not excludeShowing then
		return
	end

	if excludeShowing then
		ExcludeHide(hideOnly)
		excludeShowing = false
	else
		ExcludeShow()
		excludeShowing = true
	end
end

function GrOM.OpenExtrasPane(name)
	if currentExtras == name then
		return
	end

	local yOffset = GrOM.CloseExtrasPane()

	if name == "exclude" then
		ExcludeShow(yOffset)
		excludeShowing = true --obsolete?
	elseif name == "throttle" then
		ThrottleShow(yOffset)
	elseif name == "MVPs" then
		MVPsShow(yOffset)
	elseif name == "saves" then
		SavesShow(yOffset)
	end

	currentExtras = name
	local x, y = GrOM.GetCurrentExtrasPane()

	UIDropDownMenu_SetSelectedValue(GOMaticExtrasMenu, x)
	GOMaticExtrasMenuText:SetText(y)
end

local alreadyMoved = false

function GrOM.MoreClick(showing)
	if showing and alreadyMoved then
		return
	end

	GOMaticMore:Hide()

	GOMatic:ClearAllPoints()
	local x, y = GOMatic:GetLeft(), GOMatic:GetBottom()

	GrOM_Vars.expanded = true
	GOMaticLess:Show()
	GOMaticPing:Show()
	GOMaticOverrideRemote:Show()
	GOMaticSilence:Show()
	GOMaticSure:Show()
	GOMaticCancel:Show()
	GOMaticTemplateMenu:Show()
	GOMaticTemplateText:Show()
	GOMaticExtrasMenu:Show()

	GOMatic:SetWidth(400)

	if showing then
		alreadyMoved = true
		GOMatic:SetPoint("BOTTOMLEFT", "WorldFrame", "BOTTOMLEFT", x-167, y)
	else
		GOMatic:SetPoint("BOTTOMLEFT", "WorldFrame", "BOTTOMLEFT", x, y)
	end
end

function GrOM.LessClick()
	GOMaticMore:Show()

	GOMatic:ClearAllPoints()
	local x, y = GOMatic:GetLeft(), GOMatic:GetBottom()

	GrOM_Vars.expanded = false
	GOMaticLess:Hide()
	GOMaticPing:Hide()
	GOMaticOverrideRemote:Hide()
	GOMaticSilence:Hide()
	GOMaticSure:Hide()
	GOMaticCancel:Hide()
	GOMaticTemplateMenu:Hide()
	GOMaticTemplateText:Hide()
	GOMaticExtrasMenu:Hide()

	GOMatic:SetWidth(67)
	GOMatic:SetHeight(166)

	GOMatic:SetPoint("BOTTOMLEFT", "WorldFrame", "BOTTOMLEFT", x, y + GrOM.CloseExtrasPane())
end

function GrOM.OnCheckBoxClick(self)
	if self:GetName() == self:GetParent():GetName() .. "OverrideRemote" then
		GrOM_Vars.LocalOverridesRemote = not GrOM_Vars.LocalOverridesRemote
		cout(GrOM_Vars.LocalOverridesRemote and loc.overrideon or loc.overrideoff)
	elseif self:GetName() == self:GetParent():GetName() .. "Silence" then
		GrOM_Vars.Silent = not GrOM_Vars.Silent
		cout(loc.unsilenced)
	elseif self:GetName() == self:GetParent():GetName() .. "Sure" then
		GrOM_Vars.AlwaysSure = not GrOM_Vars.AlwaysSure
		cout(GrOM_Vars.AlwaysSure and loc.sureon or loc.sureoff)
	elseif self:GetName() == self:GetParent():GetName() .. "ExcludeSaves" then
		GrOM_Vars.excludeSaves = not GrOM_Vars.excludeSaves
	end
end

local selectedSave = nil

local function FindSave(name)
	local i = 0

	for k,_ in pairs(GrOM_Vars.SavedRaidLayouts) do
		i = i + 1
		if k == name then
			return i
		end
	end

	return false
end

function GrOM.RenameSave(name, overWrite)
	if (not overWrite) and FindSave(name) then
		GrOM.SaveName = name
		StaticPopup_Show("GOMatic_DO_RENAME", name)
		return
	end

	if (not selectedSave) or (not GrOM_Vars.SavedRaidLayouts[selectedSave]) then
		return
	end

	GrOM_Vars.SavedRaidLayouts[name] = CopyTable(GrOM_Vars.SavedRaidLayouts[selectedSave])
	GrOM_Vars.SavedRaidLayouts[selectedSave] = nil
	
	selectedSave = name

	GrOM.UpdateSavesPane()
end

function GrOM.DeleteSave()
	GrOM_Vars.SavedRaidLayouts[selectedSave] = nil

	GrOM.UpdateSavesPane()
end

local function GetFirstSave()
	for k,_ in pairs(GrOM_Vars.SavedRaidLayouts) do
		return k
	end
end

local function UpdateSaveButtons()
	if GetFirstSave() then
		GOMaticRenameSave:Enable()
		GOMaticDeleteSave:Enable()
	else
		GOMaticRenameSave:Disable()
		GOMaticDeleteSave:Disable()
	end
end

function GrOM.UpdateSavesPane()
	if not (selectedSave and FindSave(selectedSave)) then
        selectedSave = GetFirstSave()
    end

	local i = FindSave(selectedSave)
    local offset = GOMaticSavesScrollFrame.offset
    local h, maxButtons = GrOM.GetScrollFrameButtonStats(GOMaticSavesScrollFrame:GetHeight())

    if i then
        --if the selected unit is off-screen, scroll to it
        if i - 1 < offset then
            offset = i - 1
        elseif i > (offset + maxButtons) then
            offset = i - maxButtons
            if offset < 1 then
                offset = 1
            end
        end
    end

    offset = offset * h

    GrOM.FauxScrollFrame_OnVerticalScroll(GOMaticSavesScrollFrame, offset, h, GrOM.ScrollFrameUpdate)

	GrOM.UpdateButtons()

	if currentExtras ~= "saves" then
		GOMaticSavesScrollFrame:Hide()
	end
end

local selectedMVP = nil

local function FindMVP(name)
	if not name then
		return false
	end
	
	name = strlower(name)
	
	for i = 1, getn(GrOM_Vars.MVPs) do
		if strlower(GrOM_Vars.MVPs[i]) == name then
			return i
		end
	end

	return false
end

local function UpdateMVPButtons()
	local i = FindMVP(selectedMVP)

	if not i then
		GOMaticDeleteMVP:Disable()
		GOMaticDeleteAllMVPs:Disable()
		GOMaticDemoteMVP:Disable()
		GOMaticPromoteMVP:Disable()
	else
		GOMaticDeleteMVP:Enable()
		GOMaticDeleteAllMVPs:Enable()
		if i == getn(GrOM_Vars.MVPs) then
			GOMaticDemoteMVP:Disable()
		else
			GOMaticDemoteMVP:Enable()
		end

		if i == 1 then
			GOMaticPromoteMVP:Disable()
		else
			GOMaticPromoteMVP:Enable()
		end
	end
end

function GrOM.UpdateMVPPane()
    if not selectedMVP then
        selectedMVP = GrOM_Vars.MVPs[1]
    end

    local i = FindMVP(selectedMVP)
    local offset = GOMaticMVPScrollFrame.offset
    local h, maxButtons = GrOM.GetScrollFrameButtonStats(GOMaticMVPScrollFrame:GetHeight())
    --local h, maxButtons = 16, 5

    --offset = floor((offset / h) + 0.5)
    if i then
        --if the selected unit is off-screen, scroll to it
        if i - 1 < offset then
            offset = i - 1
        elseif i > (offset + maxButtons) then
            offset = i - maxButtons
            if offset < 1 then
                offset = 1
            end
        end
    end

    offset = offset * h

    GrOM.FauxScrollFrame_OnVerticalScroll(GOMaticMVPScrollFrame, offset, h, GrOM.ScrollFrameUpdate)
end

function GrOM.AddMVP(name)
	if not FindMVP(name) then
		GrOM_Vars.MVPs[getn(GrOM_Vars.MVPs) + 1] = name
	end

	selectedMVP = name
	GrOM.UpdateMVPPane()
end

function GrOM.DeleteMVP()
	local i = FindMVP(selectedMVP)

	if i then
		table.remove(GrOM_Vars.MVPs, i)
	end

	selectedMVP = nil
	GrOM.UpdateMVPPane()
end

function GrOM.ClearMVPs()
	GrOM_Vars.MVPs = {}

	selectedMVP = nil
	GrOM.UpdateMVPPane()
end

function GrOM.DemoteMVP()
	local i = FindMVP(selectedMVP)

	if i then
		local mvp = table.remove(GrOM_Vars.MVPs, i)

		if GrOM_Vars.MVPs[i+1] then
			table.insert(GrOM_Vars.MVPs, i+1, mvp)
		else
			table.insert(GrOM_Vars.MVPs, mvp)
		end
	end

	GrOM.UpdateMVPPane()
end

function GrOM.PromoteMVP()
	local i = FindMVP(selectedMVP)

	if i then
		local mvp = table.remove(GrOM_Vars.MVPs, i)

		if GrOM_Vars.MVPs[i-1] then
			table.insert(GrOM_Vars.MVPs, i-1, mvp)
		else
			table.insert(GrOM_Vars.MVPs, mvp)
		end
	end

	GrOM.UpdateMVPPane()
end

local function PopulateScrollList(name)
	local list = {}

	if name == "GOMaticMVPScrollFrame" then
		list = CopyTable(GrOM_Vars.MVPs) --check if we need to copy it
	elseif name == "GOMaticSavesScrollFrame" then
		local i = 0
		for k,_ in pairs(GrOM_Vars.SavedRaidLayouts) do
			i = i + 1
			list[i] = k
		end
	end

	return list
end

function GrOM.GetScrollFrameButtonStats(h)
	local n = math.floor(h / 16)
	local bh = math.floor((h-(16*n))/n) + 16

	return bh, n
end

function GrOM.ScrollFrameUpdate(self)
	local name = self:GetName()
	local list = PopulateScrollList(name)

	local buttonHeight, numButtons = GrOM.GetScrollFrameButtonStats(self:GetHeight())
	--local buttonHeight, numButtons = 16, 5

	local offset = self.offset

	for i = 1, numButtons do
		local buttonText = list[i + offset] or "<???>"
		_G[name..i.."DataLabel"]:SetText(buttonText)
		if self:GetName() == "GOMaticMVPScrollFrame" then
			if buttonText and selectedMVP and strlower(buttonText) == strlower(selectedMVP) then
				_G[name..i]:LockHighlight()
			else
				_G[name..i]:UnlockHighlight()
			end
		end
		if self:GetName() == "GOMaticSavesScrollFrame" then
			if buttonText == selectedSave then
				_G[name..i]:LockHighlight()
			else
				_G[name..i]:UnlockHighlight()
			end
		end
		if i + offset > getn(list) then
			_G[name..i]:Hide()
		else
			_G[name..i]:Show()
		end
	end

	if self:GetName() == "GOMaticMVPScrollFrame" then
		UpdateMVPButtons()
	end

	if self:GetName() == "GOMaticSavesScrollFrame" then
		UpdateSaveButtons()
	end

	local maxValues = getn(list)
	if numButtons >= maxValues then
		maxValues = numButtons + 1
	end

	FauxScrollFrame_Update(self, maxValues, numButtons, buttonHeight)
end

function GrOM.CreateScrollFrameButton(buttonWidth, buttonHeight, parent, anchorTo, b)
	local f = CreateFrame("BUTTON", parent:GetName()..b, parent, "GOMaticScrollFrameButtonTemplate")
	f:SetScript("OnLoad", function(self) _G[self:GetName().."DataLabel"]:SetText("test**") end)
	local relPoint = (anchorTo == parent:GetName()) and "TOPLEFT" or "BOTTOMLEFT"
	f:SetWidth(buttonWidth)
	f:SetHeight(buttonHeight)
	f:ClearAllPoints()
	f:SetPoint("TOPLEFT", anchorTo, relPoint)
	f.parentName = parent:GetName()
	f.buttonNumber = b
	f:SetScript("OnClick", function(self) GrOM.ScrollFrameButtonClick(self.parentName, self.buttonNumber, _G[self:GetName().."DataLabel"]:GetText()) end)

	return f
end

function GrOM.ScrollFrameButtonClick(parentName, buttonNumber, text)
	if parentName == "GOMaticMVPScrollFrame" then
		if selectedMVP ~= text then
			selectedMVP = text
			GrOM.UpdateMVPPane()
		end
	elseif parentName == "GOMaticSavesScrollFrame" then
		if selectedSave ~= text then
			selectedSave = text
			GrOM.UpdateSavesPane()
		end
	end
end

function GrOM.InitializeScrollFrame(self)
	local buttonHeight, numButtons = GrOM.GetScrollFrameButtonStats(self:GetHeight())
	--local buttonHeight, numButtons = 16, 5
	local buttonWidth = self:GetWidth()

	self.offset = 0

	local b = self:GetName()

	for i = 1, numButtons do
		b = GrOM.CreateScrollFrameButton(buttonWidth, buttonHeight, self, b, i)
	end

	GrOM.ScrollFrameUpdate(self)
end

function GrOM.FauxScrollFrame_OnVerticalScroll(self, offset, itemHeight, updateFunction)
	local scrollbar = getglobal(self:GetName().."ScrollBar")
	scrollbar:SetValue(offset)
	self.offset = floor((offset / itemHeight) + 0.5)
	updateFunction(self)
end